import math
import numpy as np
from ._base import DMPBase, WeightParametersMixin
from ..utils import ensure_1d_array
from ._canonical_system import canonical_system_alpha, phase


class StateFollowingDMP(WeightParametersMixin, DMPBase):
    """State following DMP (highly experimental).

    The DMP variant that is implemented here is described in [1]_.

    Attributes
    ----------
    dt_ : float
        Time difference between DMP steps. This value can be changed to adapt
        the frequency.

    References
    ----------
    .. [1] Vidakovic, J., Jerbic, B., Sekoranja, B., Svaco, M., Suligoj, F.
       (2019). Task Dependent Trajectory Learning from Multiple Demonstrations
       Using Movement Primitives. In International Conference on Robotics in
       Alpe-Adria Danube Region (RAAD) (pp. 275-282). DOI:
       10.1007/978-3-030-19648-6_32,
       https://link.springer.com/chapter/10.1007/978-3-030-19648-6_32
    """
    def __init__(self, n_dims, execution_time, dt=0.01, n_viapoints=10,
                 int_dt=0.001, alpha_y=25.0, beta_y=6.25):
        super(StateFollowingDMP, self).__init__(n_dims, n_dims)
        self.execution_time = execution_time
        self.dt_ = dt
        self.n_viapoints = n_viapoints
        self.int_dt = int_dt

        alpha_z = canonical_system_alpha(0.01, self.execution_time, 0.0)

        self.alpha_y = ensure_1d_array(alpha_y, n_dims, "alpha_y")
        self.beta_y = ensure_1d_array(beta_y, n_dims, "beta_y")

        self.forcing_term = StateFollowingForcingTerm(
            self.n_dims, self.n_viapoints, self.execution_time, 0.0, 0.1,
            alpha_z)

    def step(self, last_y, last_yd, coupling_term=None):
        assert len(last_y) == self.n_dims
        assert len(last_yd) == self.n_dims

        self.last_t = self.t
        self.t += self.dt_

        self.current_y[:], self.current_yd[:] = last_y, last_yd
        state_following_dmp_step(
            self.last_t, self.t,
            self.current_y, self.current_yd,
            self.goal_y, self.goal_yd, self.goal_ydd,
            self.start_y, self.start_yd, self.start_ydd,
            self.execution_time, 0.0,
            self.alpha_y, self.beta_y,
            forcing_term=self.forcing_term,
            coupling_term=coupling_term,
            int_dt=self.int_dt)
        return self.current_y, self.current_yd

    def open_loop(self, run_t=None, coupling_term=None):
        return state_following_dmp_open_loop(
            self.execution_time, 0.0, self.dt_, self.start_y, self.goal_y,
            self.alpha_y, self.beta_y, self.forcing_term, coupling_term, run_t,
            self.int_dt)

    def imitate(self, T, Y, regularization_coefficient=0.0,
                allow_final_velocity=False):
        raise NotImplementedError("imitation is not yet implemented")


class StateFollowingForcingTerm:
    """Defines the shape of the motion."""
    def __init__(self, n_dims, n_viapoints, goal_t, start_t, overlap, alpha_z):
        if n_viapoints <= 0:
            raise ValueError("The number of viapoints must be > 0!")
        self.n_viapoints = n_viapoints
        if start_t >= goal_t:
            raise ValueError("Goal must be chronologically after start!")
        self.goal_t = goal_t
        self.start_t = start_t
        self.overlap = overlap
        self.alpha_z = alpha_z

        self.viapoints = np.zeros((n_viapoints, n_dims))

        self._init_rbfs(n_viapoints, start_t)

    def _init_rbfs(self, n_viapoints, start_t):
        self.log_overlap = float(-math.log(self.overlap))
        self.execution_time = self.goal_t - self.start_t
        self.centers = np.empty(n_viapoints)
        self.widths = np.empty(n_viapoints)
        step = self.execution_time / self.n_viapoints
        # do first iteration outside loop because we need access to i and i - 1
        # in loop
        t = start_t
        self.centers[0] = phase(t, self.alpha_z, self.goal_t, self.start_t)
        for i in range(1, self.n_viapoints):
            # normally lower_border + i * step but lower_border is 0
            t = i * step
            self.centers[i] = phase(t, self.alpha_z, self.goal_t, self.start_t)
            # Choose width of RBF basis functions automatically so that the
            # RBF centered at one center has value overlap at the next center
            diff = self.centers[i] - self.centers[i - 1]
            self.widths[i - 1] = self.log_overlap / diff ** 2
        # Width of last Gaussian cannot be calculated, just use the same width
        # as the one before
        self.widths[self.n_viapoints - 1] = self.widths[self.n_viapoints - 2]

    def _activations(self, z, normalized):
        z = np.atleast_2d(z)  # 1 x n_steps
        squared_dist = (z - self.centers[:, np.newaxis]) ** 2
        activations = np.exp(-self.widths[:, np.newaxis] * squared_dist)
        if normalized:
            activations /= activations.sum(axis=0)
        return activations

    def __call__(self, t, int_dt=0.001):
        z = phase(t, alpha=self.alpha_z, goal_t=self.goal_t,
                  start_t=self.start_t)
        z = np.atleast_1d(z)
        return self._activations(z, normalized=True).T


def state_following_dmp_step(
        last_t, t, current_y, current_yd, goal_y, goal_yd, goal_ydd, start_y,
        start_yd, start_ydd, goal_t, start_t, alpha_y, beta_y, forcing_term,
        coupling_term=None, coupling_term_precomputed=None, int_dt=0.001):
    if start_t >= goal_t:
        raise ValueError("Goal must be chronologically after start!")

    if t <= start_t:
        return np.copy(start_y), np.copy(start_yd), np.copy(start_ydd)

    execution_time = goal_t - start_t

    current_ydd = np.empty_like(current_yd)

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is not None:
            cd, cdd = coupling_term.coupling(current_y, current_yd)
        else:
            cd, cdd = np.zeros_like(current_y), np.zeros_like(current_y)
        if coupling_term_precomputed is not None:
            cd += coupling_term_precomputed[0]
            cdd += coupling_term_precomputed[1]

        h = forcing_term(current_t).squeeze(axis=0)

        current_ydd[:] = np.sum(h[:, np.newaxis] * alpha_y * (
            beta_y * (forcing_term.viapoints - current_y)
            - 0.5 * execution_time * current_yd[np.newaxis])
                / (0.5 * execution_time) ** 2, axis=0)
        current_ydd += cdd / (0.5 * execution_time) ** 2
        current_yd += dt * current_ydd + cd / (0.5 * execution_time)
        current_y += dt * current_yd


def state_following_dmp_open_loop(
        goal_t, start_t, dt, start_y, goal_y, alpha_y, beta_y, forcing_term,
        coupling_term=None, run_t=None, int_dt=0.001):
    t = start_t
    y = np.copy(start_y)
    yd = np.zeros_like(y)
    T = [start_t]
    Y = [np.copy(y)]
    if run_t is None:
        run_t = goal_t
    while t < run_t:
        last_t = t
        t += dt
        state_following_dmp_step(
            last_t, t, y, yd,
            goal_y=goal_y, goal_yd=np.zeros_like(goal_y),
            goal_ydd=np.zeros_like(goal_y),
            start_y=start_y, start_yd=np.zeros_like(start_y),
            start_ydd=np.zeros_like(start_y),
            goal_t=goal_t, start_t=start_t, alpha_y=alpha_y, beta_y=beta_y,
            forcing_term=forcing_term, coupling_term=coupling_term,
            int_dt=int_dt)
        T.append(t)
        Y.append(np.copy(y))
    return np.asarray(T), np.asarray(Y)



"""
===================================
Plot ProMP with Multiple Via Points
===================================

This example shows how to use ProMP with multiple via points.
"""
print(__doc__)

import numpy as np
from movement_primitives.promp import ProMP, via_points
from movement_primitives.data import generate_1d_trajectory_distribution
import matplotlib.pyplot as plt


n_demos = 100
n_steps = 101
T, Y = generate_1d_trajectory_distribution(n_demos, n_steps)
promp = ProMP(n_dims=1, n_weights_per_dim=50)
promp.imitate([T] * n_demos, Y)
Y_mean = promp.mean_trajectory(T)
Y_conf = 1.96 * np.sqrt(promp.var_trajectory(T))

y_cond = np.array([0.5, -0.5, 0.0, 1.0])
y_conditional_cov = np.zeros(4)
ts = np.array([0.2, 0.5, 0.7, 1.0])
cpromp = via_points(
    promp=promp,
    y_cond=y_cond,
    y_conditional_cov=y_conditional_cov,
    ts=ts,
)
Y_cmean = cpromp.mean_trajectory(T)
Y_cconf = 1.96 * np.sqrt(cpromp.var_trajectory(T))

plt.figure(figsize=(10, 5))

ax1 = plt.subplot(121)
ax1.set_title("Training set and ProMP")
ax1.fill_between(T, (Y_mean - Y_conf).ravel(), (Y_mean + Y_conf).ravel(), color="r", alpha=0.3)
ax1.plot(T, Y_mean, c="r", lw=2, label="ProMP")
ax1.plot(T, Y[:, :, 0].T, c="k", alpha=0.1)
ax1.set_xlim((-0.05, 1.05))
ax1.set_ylim((-2.5, 3))
ax1.legend(loc="best")

ax2 = plt.subplot(122)
ax2.set_title("Conditioned ProMP")
ax2.scatter(ts, y_cond, marker="*", s=100, c="b", label="Viapoints")
ax2.fill_between(T, (Y_cmean - Y_cconf).ravel(), (Y_cmean + Y_cconf).ravel(), color="b", alpha=0.3)
ax2.plot(T, Y_cmean, c="b", lw=2, label="Conditioned ProMP")
ax2.set_xlim((-0.05, 1.05))
ax2.set_ylim((-2.5, 3))
ax2.legend(loc="best")

ax1.set_xlabel("Time $t$ [s]")
ax1.set_ylabel("Position $y$ [m]")
ax2.set_xlabel("Time $t$ [s]")
plt.tight_layout()
plt.show()



"""
============================
LASA Handwriting with ProMPs
============================

The LASA Handwriting dataset learned with ProMPs. The dataset consists of
2D handwriting motions. The first and third column of the plot represent
demonstrations and the second and fourth column show the imitated ProMPs
with 1-sigma interval.
"""
print(__doc__)


import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.data import load_lasa
from movement_primitives.promp import ProMP


def draw(T, X, idx, axes, shape_name):
    h = int(idx / width)
    w = int(idx % width) * 2
    axes[h, w].set_title(shape_name)
    axes[h, w].plot(X[:, :, 0].T, X[:, :, 1].T)

    promp = ProMP(n_weights_per_dim=30, n_dims=X.shape[2])
    promp.imitate(T, X)
    mean = promp.mean_trajectory(T[0])
    std = np.sqrt(promp.var_trajectory(T[0]))
    axes[h, w + 1].plot(mean[:, 0], mean[:, 1], c="r")
    axes[h, w + 1].plot(mean[:, 0] - std[:, 0], mean[:, 1] - std[:, 1], c="g")
    axes[h, w + 1].plot(mean[:, 0] + std[:, 0], mean[:, 1] + std[:, 1], c="g")

    axes[h, w + 1].set_xlim(axes[h, w].get_xlim())
    axes[h, w + 1].set_ylim(axes[h, w].get_ylim())
    axes[h, w].get_yaxis().set_visible(False)
    axes[h, w].get_xaxis().set_visible(False)
    axes[h, w + 1].get_yaxis().set_visible(False)
    axes[h, w + 1].get_xaxis().set_visible(False)


width = 2
height = 5

fig, axes = plt.subplots(int(height), int(width * 2))

for i in range(width * height):
    T, X, Xd, Xdd, dt, shape_name = load_lasa(i)
    draw(T, X, i, axes, shape_name)
plt.tight_layout()
plt.show()



import numpy as np
from tqdm import tqdm
from mocap.dataset_loader import load_kuka_dataset

from movement_primitives.dmp_to_state_space_distribution import propagate_weight_distribution_to_state_space
from movement_primitives.plot import plot_trajectory_in_rows, plot_distribution_in_rows
from pytransform3d import transformations as pt
import pytransform3d.visualizer as pv
import matplotlib.pyplot as plt
from movement_primitives.kinematics import Kinematics


pattern = "data/kuka/20200129_peg_in_hole/csv_processed/01_peg_in_hole_both_arms/*.csv"
#pattern = "data/kuka/20191213_carry_heavy_load/csv_processed/01_heavy_load_no_tilt_0cm_dual_arm/*.csv"
#pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/panel_450mm_counterclockwise/*.csv"

n_weights_per_dim = 5
cache_filename = "trajectories.txt"
dataset = load_kuka_dataset(pattern, verbose=1)

mvn = propagate_weight_distribution_to_state_space(
    dataset, n_weights_per_dim, cache_filename=cache_filename,
    alpha=1e-3, kappa=10.0, verbose=1)


def sample_trajectories(mvn, n_samples, n_dims):
    print("Sampling...")
    sampled_trajectories = mvn.sample(n_samples)
    n_steps = sampled_trajectories.shape[1] // n_dims
    return sampled_trajectories.reshape(n_samples, n_steps, n_dims)


########################################################
# Compare demonstrations with state space distribution #
########################################################

mean_trajectory = mvn.mean.reshape(-1, 2 * 7)
sigma = np.sqrt(np.diag(mvn.covariance).reshape(-1, 2 * 7))

n_samples = 100
n_dims = 2 * 7
sampled_trajectories = sample_trajectories(mvn, n_samples, n_dims)
n_steps = sampled_trajectories.shape[1]

# normalize training data
normalized_length_trajectories = []
from scipy.interpolate import interp1d
for _, trajectory in tqdm(dataset):
    new_trajectory = np.empty((n_steps, trajectory.shape[1]))
    for d in range(trajectory.shape[1]):
        fun = interp1d(np.arange(len(trajectory)), trajectory[:, d])
        x = np.linspace(0, 1, n_steps) * (len(trajectory) - 1)
        new_trajectory[:, d] = fun(x)
    normalized_length_trajectories.append(new_trajectory)

axes = None
for i, trajectory in enumerate(sampled_trajectories):
    axes = plot_trajectory_in_rows(trajectory, label="sampled" if i == 0 else None, color="r", alpha=0.1, axes=axes, subplot_shape=(7, 2))
for i, trajectory in enumerate(normalized_length_trajectories):
    plot_trajectory_in_rows(trajectory, label="demos" if i == 0 else None, color="orange", alpha=1, axes=axes)
plot_distribution_in_rows(mean_trajectory, sigma, label="distribution", color="b", alpha=0.2, axes=axes)
axes[0].legend()
plt.show()


################################################################
# 3D animation of sampled trajectories from state distribution #
################################################################

# IK is slow, reduce number of animated trajectories
n_animated_trajectories = 2

with open("kuka_lbr/urdf/kuka_lbr.urdf", "r") as f:
    kin = Kinematics(f.read(), mesh_path="kuka_lbr/urdf/")
right_chain = kin.create_chain(
    ["kuka_lbr_r_joint_%d" % i for i in range(1, 8)],
    "kuka_lbr", "kuka_lbr_r_tcp", verbose=0)
left_chain = kin.create_chain(
    ["kuka_lbr_l_joint_%d" % i for i in range(1, 8)],
    "kuka_lbr", "kuka_lbr_l_tcp", verbose=0)

joint_trajectories = np.empty((n_animated_trajectories, 2, n_steps, 7))
random_state = np.random.RandomState(2)
print("Inverse kinematics...")
for k in tqdm(range(n_animated_trajectories)):
    P = sampled_trajectories[k]
    H_left = np.empty((len(P), 4, 4))
    H_right = np.empty((len(P), 4, 4))
    for t in range(len(P)):
        H_left[t] = pt.transform_from_pq(P[t, :7])
        H_right[t] = pt.transform_from_pq(P[t, 7:])
    Q_left = left_chain.inverse_trajectory(H_left, np.zeros(7), random_state=random_state)
    Q_right = right_chain.inverse_trajectory(H_right, np.zeros(7), random_state=random_state)
    joint_trajectories[k, 0] = Q_left
    joint_trajectories[k, 1] = Q_right


fig = pv.figure()
fig.plot_basis(s=0.1)
graph = fig.plot_graph(kin.tm, "kuka_lbr", show_visuals=True)
for P in sampled_trajectories:
    fig.plot_trajectory(P[:, :7], s=0.05, n_frames=3)
    fig.plot_trajectory(P[:, 7:], s=0.05, n_frames=3)
fig.view_init()


def animation_callback(
        step, left_chain, right_chain, graph, joint_trajectories,
        n_samples, n_steps):
    k = step // n_steps
    t = step % n_steps
    print("Demonstration %d/%d, step %d/%d" % (k + 1, n_samples, t + 1, n_steps))
    q_left = joint_trajectories[k][0][t]
    q_right = joint_trajectories[k][1][t]
    left_chain.forward(q_left)
    right_chain.forward(q_right)
    graph.set_data()
    return graph


fig.animate(animation_callback, len(joint_trajectories) * len(joint_trajectories[0][0]), loop=True,
            fargs=(left_chain, right_chain, graph, joint_trajectories, n_samples, n_steps))
fig.show()



import numpy as np
from ._base import DMPBase, WeightParametersMixin
from ._forcing_term import ForcingTerm
from ._canonical_system import canonical_system_alpha
from ._forcing_term import phase
from ..utils import ensure_1d_array


def dmp_step_rk4(
        last_t, t, current_y, current_yd, goal_y, goal_yd, goal_ydd, start_y,
        start_yd, start_ydd, goal_t, start_t, alpha_y, beta_y, forcing_term,
        coupling_term=None, coupling_term_precomputed=None, int_dt=0.001,
        p_gain=0.0, tracking_error=0.0, smooth_scaling=False):
    """Integrate regular DMP for one step with RK4 integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (n_dims,)
        Current position. Will be modified.

    current_yd : array, shape (n_dims,)
        Current velocity. Will be modified.

    goal_y : array, shape (n_dims,)
        Goal position.

    goal_yd : array, shape (n_dims,)
        Goal velocity.

    goal_ydd : array, shape (n_dims,)
        Goal acceleration.

    start_y : array, shape (n_dims,)
        Start position.

    start_yd : array, shape (n_dims,)
        Start velocity.

    start_ydd : array, shape (n_dims,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : array, shape (n_dims,)
        Constant in transformation system.

    beta_y : array, shape (n_dims,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    coupling_term_precomputed : tuple
        A precomputed coupling term, i.e., additional velocity and
        acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    p_gain : float, optional (default: 0)
        Proportional gain for tracking error.

    tracking_error : float, optional (default: 0)
        Tracking error from last step.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.
    """
    if coupling_term is None:
        cd, cdd = np.zeros_like(current_y), np.zeros_like(current_y)
        if coupling_term_precomputed is not None:
            cd += coupling_term_precomputed[0]
            cdd += coupling_term_precomputed[1]
    else:
        cd, cdd = None, None  # will be computed in _dmp_acc()

    # RK4 (Runge-Kutta) for 2nd order differential integration
    # (faster and more accurate than Euler integration),
    # implemented following https://math.stackexchange.com/a/2023862/64116

    # precompute constants for following queries
    execution_time = goal_t - start_t
    dt = t - last_t
    dt_2 = 0.5 * dt
    T = np.array([t, t + dt_2, t + dt])
    Z = forcing_term.phase(T)
    F = forcing_term.forcing_term(Z)
    tdd = p_gain * tracking_error / dt

    C0 = current_yd
    K0 = _dmp_acc(
        current_y, C0, cdd, alpha_y, beta_y, goal_y, goal_yd, goal_ydd,
        start_y, Z[0], execution_time, F[:, 0], coupling_term, tdd,
        smooth_scaling)
    C1 = current_yd + dt_2 * K0
    K1 = _dmp_acc(
        current_y + dt_2 * C0, C1, cdd, alpha_y, beta_y, goal_y, goal_yd,
        goal_ydd, start_y, Z[1], execution_time, F[:, 1], coupling_term, tdd,
        smooth_scaling)
    C2 = current_yd + dt_2 * K1
    K2 = _dmp_acc(
        current_y + dt_2 * C1, C2, cdd, alpha_y, beta_y, goal_y, goal_yd,
        goal_ydd, start_y, Z[1], execution_time, F[:, 1], coupling_term, tdd,
        smooth_scaling)
    C3 = current_yd + dt * K2
    K3 = _dmp_acc(
        current_y + dt * C2, C3, cdd, alpha_y, beta_y, goal_y, goal_yd,
        goal_ydd, start_y, Z[2], execution_time, F[:, 2], coupling_term, tdd,
        smooth_scaling)

    current_y += dt * (current_yd + dt / 6.0 * (K0 + K1 + K2))
    current_yd += dt / 6.0 * (K0 + 2 * K1 + 2 * K2 + K3)

    if coupling_term is not None:
        cd, _ = coupling_term.coupling(current_y, current_yd)
        current_yd += cd / execution_time


def _dmp_acc(Y, V, cdd, alpha_y, beta_y, goal_y, goal_yd, goal_ydd,
             start_y, z, execution_time, f, coupling_term, tdd,
             smooth_scaling):
    """DMP acceleration.

    Parameters
    ----------
    Y : array, shape (n_dims,)
        Current state (position).

    V : array, shape (n_dims,)
        Current state derivative (velocity).

    cdd : array, shape (n_dims,)
        Coupling term acceleration.

    alpha_y : array, shape (n_dims,)
        Constant of transformation system.

    beta_y : array, shape (n_dims,)
        Constant of transformation system.

    goal_y : shape (n_dims,)
        Goal state (position).

    goal_yd : shape (n_dims,)
        Goal state derivative (velocity).

    goal_ydd : shape (n_dims,)
        Second goal state derivative (acceleration).

    start_y : array, shape (n_dims,)
        Start position.

    z : float
        Current phase.

    execution_time : float
        Time to execute the DMP.

    f : array, shape (n_dims,)
        Forcing term acceleration.

    coupling_term : object
        Coupling term object. Must have a function 'coupling' that takes as
        arguments the current position and velocity and returns a velocity and
        acceleration. (Velocity will be ignored.)

    tdd : array, shape (n_dims,)
        Acceleration correction from tracking error controller.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    Returns
    -------
    ydd : array, shape (n_dims,)
        Resulting acceleration.
    """
    if coupling_term is not None:
        _, cdd = coupling_term.coupling(Y, V)
    if smooth_scaling:
        smoothing = beta_y * (goal_y - start_y) * z
    else:
        smoothing = 0.0
    return (
        alpha_y * (
            beta_y * (goal_y - Y)
            - execution_time * V
            - smoothing
        )
        + f
        + cdd
        + tdd
    ) / execution_time ** 2


def dmp_step_euler(
        last_t, t, current_y, current_yd, goal_y, goal_yd, goal_ydd, start_y,
        start_yd, start_ydd, goal_t, start_t, alpha_y, beta_y, forcing_term,
        coupling_term=None, coupling_term_precomputed=None, int_dt=0.001,
        p_gain=0.0, tracking_error=0.0, smooth_scaling=False):
    """Integrate regular DMP for one step with Euler integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (n_dims,)
        Current position. Will be modified.

    current_yd : array, shape (n_dims,)
        Current velocity. Will be modified.

    goal_y : array, shape (n_dims,)
        Goal position.

    goal_yd : array, shape (n_dims,)
        Goal velocity.

    goal_ydd : array, shape (n_dims,)
        Goal acceleration.

    start_y : array, shape (n_dims,)
        Start position.

    start_yd : array, shape (n_dims,)
        Start velocity.

    start_ydd : array, shape (n_dims,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : array, shape (n_dims,)
        Constant in transformation system.

    beta_y : array, shape (n_dims,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    coupling_term_precomputed : tuple
        A precomputed coupling term, i.e., additional velocity and
        acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    p_gain : float, optional (default: 0)
        Proportional gain for tracking error.

    tracking_error : float, optional (default: 0)
        Tracking error from last step.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.
    """
    if start_t >= goal_t:
        raise ValueError("Goal must be chronologically after start!")

    if t <= start_t:
        return np.copy(start_y), np.copy(start_yd), np.copy(start_ydd)

    execution_time = goal_t - start_t

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is None:
            cd, cdd = np.zeros_like(current_y), np.zeros_like(current_y)
            if coupling_term_precomputed is not None:
                cd += coupling_term_precomputed[0]
                cdd += coupling_term_precomputed[1]
        else:
            cd, cdd = None, None
        z = forcing_term.phase(current_t)
        f = forcing_term.forcing_term(z).squeeze()
        tdd = p_gain * tracking_error / dt

        ydd = _dmp_acc(
            current_y, current_yd, cdd, alpha_y, beta_y, goal_y, goal_yd,
            goal_ydd, start_y, z, execution_time, f, coupling_term, tdd,
            smooth_scaling)
        current_yd += dt * ydd
        current_y += dt * current_yd

        if coupling_term is not None:
            cd, _ = coupling_term.coupling(current_y, current_yd)
            current_yd += cd / execution_time


DMP_STEP_FUNCTIONS = {
    "rk4": dmp_step_rk4,
    "euler": dmp_step_euler
}
DEFAULT_DMP_STEP_FUNCTION = "rk4"

try:
    from ..dmp_fast import dmp_step as dmp_step_euler_cython, dmp_step_rk4 as dmp_step_rk4_cython
    DMP_STEP_FUNCTIONS["euler-cython"] = dmp_step_euler_cython
    DMP_STEP_FUNCTIONS["rk4-cython"] = dmp_step_rk4_cython
    DEFAULT_DMP_STEP_FUNCTION = "rk4-cython"
except ImportError:
    pass


class DMP(WeightParametersMixin, DMPBase):
    r"""Dynamical movement primitive (DMP).

    Equation of transformation system (according to [1]_, Eq. 2.1):

    .. math::

        \ddot{y} = (\alpha_y (\beta_y (g - y) - \tau \dot{y}) + f(z) + C_t) / \tau^2

    and if smooth scaling is activated (according to [2]_):

    .. math::

        \ddot{y} = (\alpha_y (\beta_y (g - y) - \tau \dot{y} - \underline{\beta_y (g - y_0) z}) + f(z) + C_t) / \tau^2

    Parameters
    ----------
    n_dims : int
        State space dimensions.

    execution_time : float, optional (default: 1)
        Execution time of the DMP: :math:`\tau`.

    dt : float, optional (default: 0.01)
        Time difference between DMP steps: :math:`\Delta t`.

    n_weights_per_dim : int, optional (default: 10)
        Number of weights of the function approximator per dimension.

    int_dt : float, optional (default: 0.001)
        Time difference for Euler integration of transformation system.

    p_gain : float, optional (default: 0)
        Gain for proportional controller of DMP tracking error.
        The domain is [0, execution_time**2/dt].

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    alpha_y : float or array-like, shape (n_dims,), optional (default: 25.0)
        Parameter of the transformation system.

    beta_y : float, list with length n_dims, or array with shape (n_dims,), optional (default: 6.25)
        Parameter of the transformation system.

    Attributes
    ----------
    execution_time_ : float
        Execution time of the DMP.

    dt_ : float
        Time difference between DMP steps. This value can be changed to adapt
        the frequency.

    References
    ----------
    .. [1] Ijspeert, A. J., Nakanishi, J., Hoffmann, H., Pastor, P., Schaal, S.
       (2013). Dynamical Movement Primitives: Learning Attractor Models for
       Motor Behaviors. Neural Computation 25 (2), 328-373. DOI:
       10.1162/NECO_a_00393,
       https://homes.cs.washington.edu/~todorov/courses/amath579/reading/DynamicPrimitives.pdf

    .. [2] Pastor, P., Hoffmann, H., Asfour, T., Schaal, S. (2009). Learning
       and Generalization of Motor Skills by Learning from Demonstration.
       In 2009 IEEE International Conference on Robotics and Automation,
       (pp. 763-768). DOI: 10.1109/ROBOT.2009.5152385,
       https://h2t.iar.kit.edu/pdf/Pastor2009.pdf
    """
    def __init__(self, n_dims, execution_time=1.0, dt=0.01,
                 n_weights_per_dim=10, int_dt=0.001, p_gain=0.0,
                 smooth_scaling=False, alpha_y=25.0, beta_y=6.25):
        super(DMP, self).__init__(n_dims, n_dims)
        self._execution_time = execution_time
        self.dt_ = dt
        self.n_weights_per_dim = n_weights_per_dim
        self.int_dt = int_dt
        self.p_gain = p_gain
        self.smooth_scaling = smooth_scaling

        self._init_forcing_term()

        self.alpha_y = ensure_1d_array(alpha_y, n_dims, "alpha_y")
        self.beta_y = ensure_1d_array(beta_y, n_dims, "beta_y")

    def _init_forcing_term(self):
        alpha_z = canonical_system_alpha(0.01, self.execution_time_, 0.0)
        self.forcing_term = ForcingTerm(
            self.n_dims, self.n_weights_per_dim, self.execution_time_,
            0.0, 0.8, alpha_z)

    def get_execution_time_(self):
        return self._execution_time

    def set_execution_time_(self, execution_time):
        self._execution_time = execution_time
        weights = self.forcing_term.weights_
        self._init_forcing_term()
        self.forcing_term.weights_ = weights

    execution_time_ = property(get_execution_time_, set_execution_time_)

    def step(self, last_y, last_yd, coupling_term=None,
             step_function=DEFAULT_DMP_STEP_FUNCTION):
        """DMP step.

        Parameters
        ----------
        last_y : array, shape (n_dims,)
            Last state.

        last_yd : array, shape (n_dims,)
            Last time derivative of state (e.g., velocity).

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        step_function : str, optional (default: 'rk4-cython')
            DMP integration function. Possible options: 'rk4', 'euler',
            'euler-cython', 'rk4-cython'.

        Returns
        -------
        y : array, shape (n_dims,)
            Next state.

        yd : array, shape (n_dims,)
            Next time derivative of state (e.g., velocity).
        """
        try:
            step_function = DMP_STEP_FUNCTIONS[step_function]
        except KeyError:
            raise ValueError(
                f"Step function must be in {DMP_STEP_FUNCTIONS.keys()}.")

        assert len(last_y) == self.n_dims
        assert len(last_yd) == self.n_dims
        self.last_t = self.t
        self.t += self.dt_

        if not self.initialized:
            self.current_y = np.copy(self.start_y)
            self.current_yd = np.copy(self.start_yd)
            self.initialized = True

        # https://github.com/studywolf/pydmps/blob/master/pydmps/cs.py
        tracking_error = self.current_y - last_y

        step_function(
            self.last_t, self.t,
            self.current_y, self.current_yd,
            self.goal_y, self.goal_yd, self.goal_ydd,
            self.start_y, self.start_yd, self.start_ydd,
            self.execution_time_, 0.0,
            self.alpha_y, self.beta_y,
            self.forcing_term,
            coupling_term=coupling_term,
            int_dt=self.int_dt,
            p_gain=self.p_gain,
            tracking_error=tracking_error,
            smooth_scaling=self.smooth_scaling)
        return np.copy(self.current_y), np.copy(self.current_yd)

    def open_loop(self, run_t=None, coupling_term=None,
                  step_function=DEFAULT_DMP_STEP_FUNCTION):
        """Run DMP open loop.

        Parameters
        ----------
        run_t : float, optional (default: execution_time)
            Run time of DMP. Can be shorter or longer than execution_time.

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        step_function : str, optional (default: 'rk4-cython')
            DMP integration function. Possible options: 'rk4', 'euler',
            'euler-cython', 'rk4-cython'.

        Returns
        -------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, n_dims)
            State at each step.

        Raises
        ------
        ValueError
            If step function is unknown.
        """
        try:
            step_function = DMP_STEP_FUNCTIONS[step_function]
        except KeyError:
            raise ValueError(
                f"Step function must be in {DMP_STEP_FUNCTIONS.keys()}.")

        return dmp_open_loop(
            self.execution_time_, 0.0, self.dt_,
            self.start_y, self.goal_y,
            self.alpha_y, self.beta_y,
            self.forcing_term,
            coupling_term,
            run_t, self.int_dt,
            step_function,
            smooth_scaling=self.smooth_scaling)

    def imitate(self, T, Y, regularization_coefficient=0.0,
                allow_final_velocity=False):
        r"""Imitate demonstration.

        Target forces of the forcing term are computed according to

        .. math::

            f_{target} =
            \tau^2 \ddot{y}_{demo}
            - \alpha_y(
                \beta_y (g-y_{demo})
                - \tau \dot{y}_{demo}
                - \underline{\beta_y (g-y_0) z}
            ),

        where the underlined part is only used when smooth scaling is used.

        Parameters
        ----------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, n_dims)
            State at each step.

        regularization_coefficient : float, optional (default: 0)
            Regularization coefficient for regression.

        allow_final_velocity : bool, optional (default: False)
            Allow a final velocity.
        """
        self.forcing_term.weights_[:, :], start_y, _, _, goal_y, _, _ = \
            dmp_imitate(
            T, Y,
            n_weights_per_dim=self.n_weights_per_dim,
            regularization_coefficient=regularization_coefficient,
            alpha_y=self.alpha_y, beta_y=self.beta_y,
            overlap=self.forcing_term.overlap,
            alpha_z=self.forcing_term.alpha_z,
            allow_final_velocity=allow_final_velocity,
            smooth_scaling=self.smooth_scaling)
        self.configure(start_y=start_y, goal_y=goal_y)


def dmp_transformation_system(
        Y, V, alpha_y, beta_y, goal_y, goal_yd, goal_ydd, start_y, z,
        execution_time):
    """Compute acceleration generated by transformation system of DMP."""
    return (
        alpha_y * (
            beta_y * (goal_y - Y)
            + execution_time * (goal_yd - V)
            - beta_y * (goal_y - start_y) * z
        )
    ) / execution_time ** 2 + goal_ydd


def determine_forces(T, Y, alpha_y, beta_y, alpha_z, allow_final_velocity,
                     smooth_scaling=False):
    """Determine forces that the forcing term should generate.

    Parameters
    ----------
    T : array, shape (n_steps,)
        Time of each step.

    Y : array, shape (n_steps, n_dims)
        Position at each step.

    alpha_y : array, shape (n_dims,)
        Parameter of the transformation system.

    beta_y : array, shape (n_dims,)
        Parameter of the transformation system.

    alpha_z : float
        Parameter of the canonical system.

    allow_final_velocity : bool
        Whether a final velocity is allowed. Will be set to 0 otherwise.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    Returns
    -------
    F : array, shape (n_steps, n_dims)
        Forces.

    start_y : array, shape (n_dims,)
        Start position.

    start_yd : array, shape (n_dims,)
        Start velocity.

    start_ydd : array, shape (n_dims,)
        Start acceleration.

    goal_y : array, shape (n_dims,)
        Final position.

    goal_yd : array, shape (n_dims,)
        Final velocity.

    goal_ydd : array, shape (n_dims,)
        Final acceleration.
    """
    n_dims = Y.shape[1]
    DT = np.gradient(T)
    Yd = np.empty_like(Y)
    for d in range(n_dims):
        Yd[:, d] = np.gradient(Y[:, d]) / DT
    if not allow_final_velocity:
        Yd[-1, :] = 0.0
    Ydd = np.empty_like(Y)
    for d in range(n_dims):
        Ydd[:, d] = np.gradient(Yd[:, d]) / DT
    Ydd[-1, :] = 0.0

    execution_time = T[-1] - T[0]
    goal_y = Y[-1]
    start_y = Y[0]
    Z = phase(T, alpha_z, T[-1], T[0])
    F = np.empty((len(T), n_dims))
    for t in range(len(T)):
        if smooth_scaling:
            smoothing = beta_y * (goal_y - start_y) * Z[t]
        else:
            smoothing = 0.0
        F[t, :] = execution_time ** 2 * Ydd[t] - alpha_y * (
            beta_y * (goal_y - Y[t])
            - Yd[t] * execution_time
            - smoothing
        )
    return F, Y[0], Yd[0], Ydd[0], Y[-1], Yd[-1], Ydd[-1]


def dmp_imitate(
        T, Y, n_weights_per_dim, regularization_coefficient, alpha_y, beta_y,
        overlap, alpha_z, allow_final_velocity,
        determine_forces=determine_forces, smooth_scaling=False):
    """Compute weights and metaparameters of DMP.

    Parameters
    ----------
    T : array, shape (n_steps,)
        Time of each step.

    Y : array, shape (n_steps, n_dims)
        Position at each step.

    n_weights_per_dim : int
        Number of weights per dimension.

    regularization_coefficient : float, optional (default: 0)
        Regularization coefficient for regression.

    alpha_y : array, shape (n_dims,)
        Parameter of the transformation system.

    beta_y : array, shape (n_dims,)
        Parameter of the transformation system.

    overlap : float
        At which value should radial basis functions of the forcing term
        overlap?

    alpha_z : float
        Parameter of the canonical system.

    allow_final_velocity : bool
        Whether a final velocity is allowed. Will be set to 0 otherwise.

    determine_forces : callable
        Function to compute forces of the forcing term and metaparameters given
        the demonstration.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    Returns
    -------
    weights : array, shape (n_dims, n_weights_per_dim)
        Weights of the forcing term.

    start_y : array, shape (n_dims,)
        Start position.

    start_yd : array, shape (n_dims,)
        Start velocity.

    start_ydd : array, shape (n_dims,)
        Start acceleration.

    goal_y : array, shape (n_dims,)
        Final position.

    goal_yd : array, shape (n_dims,)
        Final velocity.

    goal_ydd : array, shape (n_dims,)
        Final acceleration.
    """
    if regularization_coefficient < 0.0:
        raise ValueError("Regularization coefficient must be >= 0!")

    forcing_term = ForcingTerm(
        Y.shape[1], n_weights_per_dim, T[-1], T[0], overlap, alpha_z)
    F, start_y, start_yd, start_ydd, goal_y, goal_yd, goal_ydd = \
        determine_forces(
            T, Y, alpha_y, beta_y,
            alpha_z, allow_final_velocity,
            smooth_scaling
        )
    # F shape (n_steps, n_dims)

    X = forcing_term.design_matrix(T)  # shape (n_weights_per_dim, n_steps)

    return (ridge_regression(X, F, regularization_coefficient),
            start_y, start_yd, start_ydd, goal_y, goal_yd, goal_ydd)


def ridge_regression(X, Y, regularization_coefficient):
    r"""Ridge regression: linear regression with Tikhonov regularization.

    A linear model :math:`Y = X W^T` or :math:`y = W x` that minimizes
    :math:`||Y - X W^T||_2^2 + \lambda ||W||_2^2`, where :math:`\lambda`
    is the regularization coefficient.

    Parameters
    ----------
    X : array, shape (n_samples, n_features)
        Input features of linear model.

    Y : array, shape (n_samples, n_outputs)
        Target values.

    regularization_coefficient : float
        Regularization coefficient.

    Returns
    -------
    weights : array, shape (n_outputs, n_features)
        Weights of linear model.
    """
    return np.linalg.pinv(
        X.dot(X.T) + regularization_coefficient * np.eye(X.shape[0])
    ).dot(X).dot(Y).T


def dmp_open_loop(
        goal_t, start_t, dt, start_y, goal_y, alpha_y, beta_y, forcing_term,
        coupling_term=None, run_t=None, int_dt=0.001,
        step_function=dmp_step_rk4, smooth_scaling=False,
        start_yd=None, start_ydd=None, goal_yd=None, goal_ydd=None):
    """Run DMP without external feedback.

    Parameters
    ----------
    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    dt : float, optional (default: 0.01)
        Time difference between DMP steps.

    start_y : array, shape (n_dims,)
        Start position.

    goal_y : array, shape (n_dims,)
        Goal position.

    alpha_y : array, shape (n_dims,)
        Constant in transformation system.

    beta_y : array, shape (n_dims,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    run_t : float, optional (default: goal_t)
        Time at which the DMP will be stopped.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    step_function : callable, optional (default: dmp_step_rk4)
        DMP integration function.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    start_yd : shape (n_dims,)
        Start state derivative (velocity).

    start_ydd : shape (n_dims,)
        Second start state derivative (acceleration).

    goal_yd : shape (n_dims,)
        Goal state derivative (velocity).

    goal_ydd : shape (n_dims,)
        Second goal state derivative (acceleration).

    Returns
    -------
    T : array, shape (n_steps,)
        Times.

    Y : array, shape (n_steps, n_dims)
        Positions.
    """
    if goal_yd is None:
        goal_yd = np.zeros_like(goal_y)
    if goal_ydd is None:
        goal_ydd = np.zeros_like(goal_y)
    if start_yd is None:
        start_yd = np.zeros_like(start_y)
    if start_ydd is None:
        start_ydd = np.zeros_like(start_y)

    if run_t is None:
        run_t = goal_t

    current_y = np.copy(start_y)
    current_yd = np.copy(start_yd)

    T = np.arange(start_t, run_t + dt, dt)
    Y = np.empty((len(T), len(current_y)))
    Y[0] = current_y

    for i in range(1, len(T)):
        step_function(
            T[i - 1], T[i], current_y, current_yd,
            goal_y=goal_y, goal_yd=goal_yd, goal_ydd=goal_ydd,
            start_y=start_y, start_yd=start_yd, start_ydd=start_ydd,
            goal_t=goal_t, start_t=start_t,
            alpha_y=alpha_y, beta_y=beta_y,
            forcing_term=forcing_term, coupling_term=coupling_term,
            int_dt=int_dt, smooth_scaling=smooth_scaling)
        Y[i] = current_y

    return T, Y



"""
======================
Spatial Scaling of DMP
======================

The standard DMP definition of Ijspeert et al. (2013) does not scale well,
when the original demonstration has both the start and the goal state close
together: small perturbation in the start and/or goal state often result in
large accelerations. To fix this issue, our implementation of the DMP does
not scale the forcing term by $(g - y_0)$. We also adopted the modification of

P. Pastor, H. Hoffmann, T. Asfour, S. Schaal:
Learning and Generalization of Motor Skills by Learning from Demonstration

Note that this has to be activated explicitly by setting smooth scaling in the
constructor of the DMP. This behavior is implemented for all types of DMPs.
This example demonstrates the behavior of this DMP implementation in these
cases.
"""
import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.dmp import DMP


T = np.linspace(0, 1, 101)
x = np.sin(T ** 2 * 1.99 * np.pi)
y = np.cos(T ** 2 * 1.99 * np.pi)
z = qx = qy = qz = np.zeros_like(x)
qw = np.ones_like(x)
Y = np.column_stack((x, y, z, qw, qx, qy, qz, x, y, z, qw, qx, qy, qz))
start = Y[0]
goal = Y[-1]
new_start = np.array([0.5, 0.5, 0, 1, 0, 0, 0, 0.5, 0.5, 0, 1, 0, 0, 0])
new_goal = np.array([-0.5, 0.5, 0, 1, 0, 0, 0, -0.5, 0.5, 0, 1, 0, 0, 0])
Y_start_shifted = Y - start[np.newaxis] + new_start[np.newaxis]
Y_goal_shifted = Y - goal[np.newaxis] + new_goal[np.newaxis]

dmp = DMP(n_dims=len(start), execution_time=1.0, dt=0.01, n_weights_per_dim=20,
          smooth_scaling=True)
dmp.imitate(T, Y)
dmp.configure(start_y=new_start, goal_y=new_goal)
_, Y_dmp = dmp.open_loop()

plt.plot(Y[:, 0], Y[:, 1], label=r"Demonstration, $g \approx y_0$", ls="--")
plt.plot(Y_start_shifted[:, 0], Y_start_shifted[:, 1], label="Original shape with new start", ls="--")
plt.plot(Y_goal_shifted[:, 0], Y_goal_shifted[:, 1], label="Original shape with new goal", ls="--")
plt.plot(Y_dmp[:, 0], Y_dmp[:, 1], label="DMP with new goal", lw=5, alpha=0.5)
plt.scatter(Y_dmp[:, 0], Y_dmp[:, 1], alpha=0.5)
plt.scatter(goal[0], goal[1], c="r", label="Goal of demonstration: $g$")
plt.scatter(start[0], start[1], c="g", label="Start of demonstration: $y_0$")
plt.scatter(new_start[0], new_start[1], c="c", label="New start: $y_0'$")
plt.scatter(new_goal[0], new_goal[1], c="y", label="New goal: $g'$")
plt.xlabel("$y_1$")
plt.ylabel("$y_2$")
plt.legend(loc="best", ncol=2)
plt.xlim((-1.8, 2.1))
plt.ylim((-1.7, 2.2))
plt.tight_layout()
plt.show()



"""Base Classes
============

Base classes of movement primitives."""
import abc
import numpy as np
from .utils import check_1d_array_length


class PointToPointMovement(abc.ABC):
    """Base class for point to point movements (discrete motions).

    Parameters
    ----------
    n_pos_dims : int
        Number of dimensions of the position that will be controlled.

    n_vel_dims : int
        Number of dimensions of the velocity that will be controlled.

    Attributes
    ----------
    n_dims : int
        Number of state space dimensions.

    n_vel_dims : int
        Number of velocity dimensions.

    t : float
        Current time.

    last_t : float
        Time during last step.

    start_y : array, shape (n_dims,)
        Initial state.

    start_yd : array, shape (n_vel_dims,)
        Initial velocity.

    start_ydd : array, shape (n_vel_dims,)
        Initial acceleration.

    goal_y : array, shape (n_dims,)
        Goal state.

    goal_yd : array, shape (n_vel_dims,)
        Goal velocity.

    goal_ydd : array, shape (n_vel_dims,)
        Goal acceleration.

    current_y : array, shape (n_dims,)
        Current state.

    current_yd : array, shape (n_vel_dims,)
        Current velocity.
    """
    def __init__(self, n_pos_dims, n_vel_dims):
        self.n_dims = n_pos_dims
        self.n_vel_dims = n_vel_dims

        self.t = 0.0
        self.last_t = None

        self.start_y = np.zeros(n_pos_dims)
        self.start_yd = np.zeros(n_vel_dims)
        self.start_ydd = np.zeros(n_vel_dims)

        self.goal_y = np.zeros(n_pos_dims)
        self.goal_yd = np.zeros(n_vel_dims)
        self.goal_ydd = np.zeros(n_vel_dims)

        self.current_y = np.zeros(n_pos_dims)
        self.current_yd = np.zeros(n_vel_dims)

    def configure(
            self, t=None, start_y=None, start_yd=None, start_ydd=None,
            goal_y=None, goal_yd=None, goal_ydd=None):
        """Set meta parameters.

        Parameters
        ----------
        t : float, optional
            Time at current step.

        start_y : array, shape (n_dims,)
            Initial state.

        start_yd : array, shape (n_vel_dims,)
            Initial velocity.

        start_ydd : array, shape (n_vel_dims,)
            Initial acceleration.

        goal_y : array, shape (n_dims,)
            Goal state.

        goal_yd : array, shape (n_vel_dims,)
            Goal velocity.

        goal_ydd : array, shape (n_vel_dims,)
            Goal acceleration.

        Raises
        ------
        ValueError
            If the length of the configured meta parameter is not correct.
        """
        if t is not None:
            self.t = t

        if start_y is not None:
            check_1d_array_length(start_y, "start_y", self.n_dims)
            self.start_y = start_y
        if start_yd is not None:
            check_1d_array_length(start_yd, "start_yd", self.n_vel_dims)
            self.start_yd = start_yd
        if start_ydd is not None:
            check_1d_array_length(start_ydd, "start_ydd", self.n_vel_dims)
            self.start_ydd = start_ydd
        if goal_y is not None:
            check_1d_array_length(goal_y, "goal_y", self.n_dims)
            self.goal_y = goal_y
        if goal_yd is not None:
            check_1d_array_length(goal_yd, "goal_yd", self.n_vel_dims)
            self.goal_yd = goal_yd
        if goal_ydd is not None:
            check_1d_array_length(goal_ydd, "goal_ydd", self.n_vel_dims)
            self.goal_ydd = goal_ydd

    @abc.abstractmethod
    def step(self, last_y, last_yd):
        """Perform step.

        Parameters
        ----------
        last_y : array, shape (n_dims,)
            Last state.

        last_yd : array, shape (n_dims,)
            Last time derivative of state (e.g., velocity).

        Returns
        -------
        y : array, shape (n_dims,)
            Next state.

        yd : array, shape (n_dims,)
            Next time derivative of state (e.g., velocity).
        """

    def n_steps_open_loop(self, last_y, last_yd, n_steps):
        """Perform 'n_steps' steps.

        Parameters
        ----------
        last_y : array, shape (n_dims,)
            Last state.

        last_yd : array, shape (n_dims,)
            Last time derivative of state (e.g., velocity).

        n_steps : int
            Number of steps.

        Returns
        -------
        y : array, shape (n_dims,)
            Next state.

        yd : array, shape (n_dims,)
            Next time derivative of state (e.g., velocity).
        """
        for _ in range(n_steps):
            last_y, last_yd = self.step(last_y, last_yd)
        return last_y, last_yd



import numpy as np
import pytransform3d.rotations as pr
from ._base import DMPBase, WeightParametersMixin
from ..utils import ensure_1d_array
from ._canonical_system import canonical_system_alpha
from ._forcing_term import ForcingTerm, phase
from ._dmp import dmp_imitate
from ._cartesian_dmp import dmp_quaternion_imitation


pps = [0, 1, 2, 7, 8, 9]
pvs = [0, 1, 2, 6, 7, 8]


def dmp_step_dual_cartesian_python(
        last_t, t,
        current_y, current_yd,
        goal_y, goal_yd, goal_ydd,
        start_y, start_yd, start_ydd,
        goal_t, start_t, alpha_y, beta_y,
        forcing_term, coupling_term=None, int_dt=0.001,
        p_gain=0.0, tracking_error=None, smooth_scaling=False):
    """Integrate bimanual Cartesian DMP for one step with Euler integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (14,)
        Current position. Will be modified.

    current_yd : array, shape (12,)
        Current velocity. Will be modified.

    goal_y : array, shape (14,)
        Goal position.

    goal_yd : array, shape (12,)
        Goal velocity.

    goal_ydd : array, shape (12,)
        Goal acceleration.

    start_y : array, shape (14,)
        Start position.

    start_yd : array, shape (12,)
        Start velocity.

    start_ydd : array, shape (12,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : float or array with shape (12,), optional (default: 25.0)
        Constant in transformation system.

    beta_y : float or array with shape (12,), optional (default: 6.25)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    p_gain : float, optional (default: 0)
        Proportional gain for tracking error.

    tracking_error : float, optional (default: 0)
        Tracking error from last step.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.
    """
    if t <= start_t:
        current_y[:] = start_y
        current_yd[:] = start_yd

    execution_time = goal_t - start_t

    current_ydd = np.empty_like(current_yd)

    cd, cdd = np.zeros_like(current_yd), np.zeros_like(current_ydd)

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is not None:
            cd[:], cdd[:] = coupling_term.coupling(current_y, current_yd)

        z = forcing_term.phase(current_t)
        f = forcing_term.forcing_term(z).squeeze()
        if tracking_error is not None:
            cdd[pvs] += p_gain * tracking_error[pps] / dt
            for ops, ovs in ((slice(3, 7), slice(3, 6)),
                             (slice(10, 14), slice(9, 12))):
                cdd[ovs] += p_gain * pr.compact_axis_angle_from_quaternion(
                    tracking_error[ops]) / dt

        if smooth_scaling:
            smoothing = beta_y[pps] * (goal_y[pps] - start_y[pps]) * z
        else:
            smoothing = 0.0

        # position components
        current_ydd[pvs] = (
            alpha_y[pvs] * (
                beta_y[pvs] * (goal_y[pps] - current_y[pps])
                - execution_time * current_yd[pvs]
                - smoothing
            )
            + f[pvs]
            + cdd[pvs]
        ) / execution_time ** 2
        current_yd[pvs] += dt * current_ydd[pvs] + cd[pvs] / execution_time
        current_y[pps] += dt * current_yd[pvs]

        # orientation components
        for ops, ovs in ((slice(3, 7), slice(3, 6)),
                         (slice(10, 14), slice(9, 12))):
            if smooth_scaling:
                goal_y_minus_start_y = pr.compact_axis_angle_from_quaternion(
                    pr.concatenate_quaternions(goal_y[ops], pr.q_conj(start_y[ops])))
                smoothing = beta_y[ovs] * z * goal_y_minus_start_y
            else:
                smoothing = 0.0
            current_ydd[ovs] = (
                alpha_y[ovs] * (
                    beta_y[ovs] * pr.compact_axis_angle_from_quaternion(pr.concatenate_quaternions(
                        goal_y[ops], pr.q_conj(current_y[ops])))
                    - execution_time * current_yd[ovs]
                    - smoothing
                )
                + f[ovs]
                + cdd[ovs]
            ) / execution_time ** 2
            current_yd[ovs] += dt * current_ydd[ovs] + cd[ovs] / execution_time
            current_y[ops] = pr.concatenate_quaternions(
                pr.quaternion_from_compact_axis_angle(dt * current_yd[ovs]), current_y[ops])


DUAL_CARTESIAN_DMP_STEP_FUNCTIONS = {
    "python": dmp_step_dual_cartesian_python
}


try:
    from ..dmp_fast import dmp_step_dual_cartesian
    DUAL_CARTESIAN_DMP_STEP_FUNCTIONS["cython"] = dmp_step_dual_cartesian
    DEFAULT_DUAL_CARTESIAN_DMP_STEP_FUNCTION = "cython"
except ImportError:
    DEFAULT_DUAL_CARTESIAN_DMP_STEP_FUNCTION = "python"


class DualCartesianDMP(WeightParametersMixin, DMPBase):
    """Dual cartesian dynamical movement primitive.

    Each of the two Cartesian DMPs handles orientation and position separately.
    The orientation is represented by a quaternion.
    See :class:`CartesianDMP` for details about the equation of the
    transformation system.

    While the dimension of the state space is 14, the dimension of the
    velocity, acceleration, and forcing term is 12.

    Parameters
    ----------
    execution_time : float, optional (default: 1)
        Execution time of the DMP.

    dt : float, optional (default: 0.01)
        Time difference between DMP steps.

    n_weights_per_dim : int, optional (default: 10)
        Number of weights of the function approximator per dimension.

    int_dt : float, optional (default: 0.001)
        Time difference for Euler integration.

    p_gain : float, optional (default: 0)
        Gain for proportional controller of DMP tracking error.
        The domain is [0, execution_time**2/dt].

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    alpha_y : float or array-like, shape (12,), optional (default: 25.0)
        Parameter of the transformation system.

    beta_y : float or array-like, shape (12,), optional (default: 6.25)
        Parameter of the transformation system.

    Attributes
    ----------
    execution_time_ : float
        Execution time of the DMP.

    dt_ : float
        Time difference between DMP steps. This value can be changed to adapt
        the frequency.

    References
    ----------
    .. [1] Ude, A., Nemec, B., Petric, T., Morimoto, J. (2014).
       Orientation in Cartesian space dynamic movement primitives.
       In IEEE International Conference on Robotics and Automation (ICRA)
       (pp. 2997-3004). DOI: 10.1109/ICRA.2014.6907291,
       https://acat-project.eu/modules/BibtexModule/uploads/PDF/udenemecpetric2014.pdf

    .. [2] Pastor, P., Hoffmann, H., Asfour, T., Schaal, S. (2009). Learning
       and Generalization of Motor Skills by Learning from Demonstration.
       In 2009 IEEE International Conference on Robotics and Automation,
       (pp. 763-768). DOI: 10.1109/ROBOT.2009.5152385,
       https://h2t.iar.kit.edu/pdf/Pastor2009.pdf
    """
    def __init__(self, execution_time=1.0, dt=0.01, n_weights_per_dim=10,
                 int_dt=0.001, p_gain=0.0, smooth_scaling=False, alpha_y=25.0, beta_y=6.25):
        super(DualCartesianDMP, self).__init__(14, 12)
        self._execution_time = execution_time
        self.dt_ = dt
        self.n_weights_per_dim = n_weights_per_dim
        self.int_dt = int_dt
        self.p_gain = p_gain
        self.smooth_scaling = smooth_scaling

        self._init_forcing_term()

        self.alpha_y = ensure_1d_array(alpha_y, 12, "alpha_y")
        self.beta_y = ensure_1d_array(beta_y, 12, "beta_y")

    def _init_forcing_term(self):
        alpha_z = canonical_system_alpha(0.01, self.execution_time_, 0.0)
        self.forcing_term = ForcingTerm(
            12, self.n_weights_per_dim, self.execution_time_, 0.0, 0.8,
            alpha_z)

    def get_execution_time_(self):
        return self._execution_time

    def set_execution_time_(self, execution_time):
        self._execution_time = execution_time
        weights = self.forcing_term.weights_
        self._init_forcing_term()
        self.forcing_term.weights_ = weights

    execution_time_ = property(get_execution_time_, set_execution_time_)

    def step(self, last_y, last_yd, coupling_term=None,
             step_function=DUAL_CARTESIAN_DMP_STEP_FUNCTIONS[
                 DEFAULT_DUAL_CARTESIAN_DMP_STEP_FUNCTION]):
        """DMP step.

        Parameters
        ----------
        last_y : array, shape (14,)
            Last state.

        last_yd : array, shape (12,)
            Last time derivative of state (velocity).

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        step_function : callable, optional (default: cython code if available)
            DMP integration function.

        Returns
        -------
        y : array, shape (14,)
            Next state.

        yd : array, shape (12,)
            Next time derivative of state (velocity).
        """
        assert len(last_y) == self.n_dims
        assert len(last_yd) == 12

        self.last_t = self.t
        self.t += self.dt_

        if not self.initialized:
            self.current_y = np.copy(self.start_y)
            self.current_yd = np.copy(self.start_yd)
            self.initialized = True

        tracking_error = self.current_y - last_y
        for ops in (slice(3, 7), slice(10, 14)):
            tracking_error[ops] = pr.concatenate_quaternions(
                self.current_y[ops], pr.q_conj(last_y[ops]))
        self.current_y[:], self.current_yd[:] = last_y, last_yd
        step_function(
            self.last_t, self.t, self.current_y, self.current_yd,
            self.goal_y, self.goal_yd, self.goal_ydd,
            self.start_y, self.start_yd, self.start_ydd,
            self.execution_time_, 0.0,
            self.alpha_y, self.beta_y,
            self.forcing_term, coupling_term,
            self.int_dt,
            self.p_gain, tracking_error,
            self.smooth_scaling)

        return np.copy(self.current_y), np.copy(self.current_yd)

    def open_loop(self, run_t=None, coupling_term=None,
                  step_function=DEFAULT_DUAL_CARTESIAN_DMP_STEP_FUNCTION):
        """Run DMP open loop.

        Parameters
        ----------
        run_t : float, optional (default: execution_time)
            Run time of DMP. Can be shorter or longer than execution_time.

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        step_function : str, optional (default: 'cython' if available)
            DMP integration function. Possible options: 'python', 'cython'.

        Returns
        -------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, 14)
            State at each step.
        """
        try:
            step_function = DUAL_CARTESIAN_DMP_STEP_FUNCTIONS[step_function]
        except KeyError:
            raise ValueError(
                f"Step function must be in "
                f"{DUAL_CARTESIAN_DMP_STEP_FUNCTIONS.keys()}.")

        if run_t is None:
            run_t = self.execution_time_
        self.t = 0.0
        T = [self.t]
        Y = [np.copy(self.start_y)]
        y = np.copy(self.start_y)
        yd = np.copy(self.start_yd)
        while self.t < run_t:
            y, yd = self.step(y, yd, coupling_term, step_function)
            T.append(self.t)
            Y.append(np.copy(self.current_y))
        self.t = 0.0
        return np.array(T), np.vstack(Y)

    def imitate(self, T, Y, regularization_coefficient=0.0,
                allow_final_velocity=False):
        """Imitate demonstration.

        Target forces for the forcing term are computed for the positions
        in a similar way as in :func:`DMP.imitate`. For the orientations
        we adapt this to handle quaternions adequately.

        Parameters
        ----------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, 14)
            State at each step.

        regularization_coefficient : float, optional (default: 0)
            Regularization coefficient for regression.

        allow_final_velocity : bool, optional (default: False)
            Allow a final velocity.
        """
        self.forcing_term.weights_[:3, :] = dmp_imitate(
            T, Y[:, :3],
            n_weights_per_dim=self.n_weights_per_dim,
            regularization_coefficient=regularization_coefficient,
            alpha_y=self.alpha_y[:3], beta_y=self.beta_y[:3],
            overlap=self.forcing_term.overlap,
            alpha_z=self.forcing_term.alpha_z,
            allow_final_velocity=allow_final_velocity,
            smooth_scaling=self.smooth_scaling)[0]
        self.forcing_term.weights_[3:6, :] = dmp_quaternion_imitation(
            T, Y[:, 3:7],
            n_weights_per_dim=self.n_weights_per_dim,
            regularization_coefficient=regularization_coefficient,
            alpha_y=self.alpha_y[3:6], beta_y=self.beta_y[3:6],
            overlap=self.forcing_term.overlap,
            alpha_z=self.forcing_term.alpha_z,
            allow_final_velocity=allow_final_velocity,
            smooth_scaling=self.smooth_scaling)[0]
        self.forcing_term.weights_[6:9, :] = dmp_imitate(
            T, Y[:, 7:10],
            n_weights_per_dim=self.n_weights_per_dim,
            regularization_coefficient=regularization_coefficient,
            alpha_y=self.alpha_y[6:9], beta_y=self.beta_y[6:9],
            overlap=self.forcing_term.overlap,
            alpha_z=self.forcing_term.alpha_z,
            allow_final_velocity=allow_final_velocity,
            smooth_scaling=self.smooth_scaling)[0]
        self.forcing_term.weights_[9:12, :] = dmp_quaternion_imitation(
            T, Y[:, 10:14],
            n_weights_per_dim=self.n_weights_per_dim,
            regularization_coefficient=regularization_coefficient,
            alpha_y=self.alpha_y[9:12], beta_y=self.beta_y[9:12],
            overlap=self.forcing_term.overlap,
            alpha_z=self.forcing_term.alpha_z,
            allow_final_velocity=allow_final_velocity,
            smooth_scaling=self.smooth_scaling)[0]

        self.configure(start_y=Y[0], goal_y=Y[-1])



"""Spring-Damper Based Attractors
==============================

Spring-damper based attractors are the basis of a DMP's transformation system.
"""
import numpy as np
import pytransform3d.rotations as pr
from .base import PointToPointMovement


class SpringDamper(PointToPointMovement):
    """Spring-damper system.

    This is similar to a DMP without the forcing term.

    Parameters
    ----------
    n_dims : int
        State space dimensions.

    dt : float, optional (default: 0.01)
        Time difference between DMP steps.

    k : float, optional (default: 1)
        Spring constant.

    c : float, optional (default: 2 * sqrt(k) (critical damping))
        Damping coefficient.

    int_dt : float, optional (default: 0.001)
        Time difference for Euler integration.
    """
    def __init__(self, n_dims, dt=0.01, k=1.0, c=None, int_dt=0.001):
        super(SpringDamper, self).__init__(n_dims, n_dims)
        self.n_dims = n_dims
        self.dt = dt
        self.k = k
        self.c = c
        self.int_dt = int_dt

        self.initialized = False
        self.configure()

    def reset(self):
        """Reset time."""
        self.t = 0.0
        self.last_t = None

    def step(self, last_y, last_yd, coupling_term=None):
        """Perform step.

        Parameters
        ----------
        last_y : array, shape (n_dims,)
            Last state.

        last_yd : array, shape (n_dims,)
            Last time derivative of state (e.g., velocity).

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        Returns
        -------
        y : array, shape (n_dims,)
            Next state.

        yd : array, shape (n_dims,)
            Next time derivative of state (e.g., velocity).
        """
        self.last_t = self.t
        self.t += self.dt

        if not self.initialized:
            self.current_y = np.copy(self.start_y)
            self.current_yd = np.copy(self.start_yd)
            self.initialized = True

        self.current_y[:], self.current_yd[:] = last_y, last_yd
        spring_damper_step(
            self.last_t, self.t,
            self.current_y, self.current_yd,
            self.goal_y,
            self.k, self.c,
            coupling_term=coupling_term,
            int_dt=self.int_dt)
        return np.copy(self.current_y), np.copy(self.current_yd)

    def open_loop(self, run_t=1.0, coupling_term=None):
        """Run open loop.

        Parameters
        ----------
        run_t : float, optional (default: execution_time)
            Run time. Can be shorter or longer than execution_time.

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        Returns
        -------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, n_dims)
            State at each step.
        """
        return spring_damper_open_loop(
            self.dt,
            self.start_y, self.goal_y,
            self.k, self.c,
            coupling_term,
            run_t, self.int_dt)


class SpringDamperOrientation(PointToPointMovement):
    """Spring-damper system for quaternions.

    This is similar to a Quaternion DMP without the forcing term.

    Parameters
    ----------
    dt : float, optional (default: 0.01)
        Time difference between DMP steps.

    k : float, optional (default: 1)
        Spring constant.

    c : float, optional (default: 2 * sqrt(k) (critical damping))
        Damping coefficient.

    int_dt : float, optional (default: 0.001)
        Time difference for Euler integration.
    """
    def __init__(self, dt=0.01, k=1.0, c=None, int_dt=0.001):
        super(SpringDamperOrientation, self).__init__(4, 3)

        self.dt = dt
        self.k = k
        self.c = c
        self.int_dt = int_dt

        self.initialized = False
        self.configure()

    def reset(self):
        """Reset time."""
        self.t = 0.0
        self.last_t = None

    def step(self, last_y, last_yd, coupling_term=None):
        """Perform step.

        Parameters
        ----------
        last_y : array, shape (n_dims,)
            Last state.

        last_yd : array, shape (n_dims,)
            Last time derivative of state (e.g., velocity).

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        Returns
        -------
        y : array, shape (n_dims,)
            Next state.

        yd : array, shape (n_dims,)
            Next time derivative of state (e.g., velocity).
        """
        self.last_t = self.t
        self.t += self.dt

        if not self.initialized:
            self.current_y = np.copy(self.start_y)
            self.current_yd = np.copy(self.start_yd)
            self.initialized = True

        self.current_y[:], self.current_yd[:] = last_y, last_yd
        spring_damper_step_quaternion(
            self.last_t, self.t,
            self.current_y, self.current_yd,
            self.goal_y,
            self.k, self.c,
            coupling_term=coupling_term,
            int_dt=self.int_dt)
        return np.copy(self.current_y), np.copy(self.current_yd)

    def open_loop(self, run_t=1.0, coupling_term=None):
        """Run open loop.

        Parameters
        ----------
        run_t : float, optional (default: execution_time)
            Run time. Can be shorter or longer than execution_time.

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        Returns
        -------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, n_dims)
            State at each step.
        """
        return spring_damper_open_loop_quaternion(
            self.dt,
            self.start_y, self.goal_y,
            self.k, self.c,
            coupling_term,
            run_t, self.int_dt)


def spring_damper_step(
        last_t, t, current_y, current_yd, goal_y, k=1.0, c=None,
        coupling_term=None, coupling_term_precomputed=None, int_dt=0.001):
    c = _default_critically_damped_c(c, k)

    current_ydd = np.empty_like(current_yd)

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is not None:
            cd, cdd = coupling_term.coupling(current_y, current_yd)
        else:
            cd, cdd = np.zeros_like(current_y), np.zeros_like(current_y)
        if coupling_term_precomputed is not None:
            cd += coupling_term_precomputed[0]
            cdd += coupling_term_precomputed[1]

        current_ydd[:] = k * (goal_y - current_y) - c * current_yd + cdd
        current_yd += dt * current_ydd + cd
        current_y += dt * current_yd


def _default_critically_damped_c(c, k):
    if c is None:
        c = 2.0 * np.sqrt(k)
    return c


def spring_damper_step_quaternion(
        last_t, t, current_y, current_yd, goal_y, k=1.0, c=None,
        coupling_term=None, coupling_term_precomputed=None, int_dt=0.001):
    c = _default_critically_damped_c(c, k)

    current_ydd = np.empty_like(current_yd)

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is not None:
            cd, cdd = coupling_term.coupling(current_y, current_yd)
        else:
            cd, cdd = np.zeros(3), np.zeros(3)
        if coupling_term_precomputed is not None:
            cd += coupling_term_precomputed[0]
            cdd += coupling_term_precomputed[1]

        current_ydd[:] = (
            k * pr.compact_axis_angle_from_quaternion(
                pr.concatenate_quaternions(goal_y, pr.q_conj(current_y)))
            - c * current_yd
            + cdd)
        current_yd += dt * current_ydd + cd
        current_y[:] = pr.concatenate_quaternions(
            pr.quaternion_from_compact_axis_angle(dt * current_yd), current_y)


def spring_damper_open_loop(
        dt, start_y, goal_y, k=1.0, c=None, coupling_term=None, run_t=1.0,
        int_dt=0.001):
    y = np.copy(start_y)
    yd = np.zeros_like(start_y)

    T = np.arange(0.0, run_t + dt, dt)
    Y = np.empty((len(T), len(y)))
    Y[0] = y
    for i in range(1, len(T)):
        spring_damper_step(
            T[i - 1], T[i], y, yd, goal_y=goal_y, k=k, c=c,
            coupling_term=coupling_term, int_dt=int_dt)
        Y[i] = y
    return T, Y


def spring_damper_open_loop_quaternion(
        dt, start_y, goal_y, k=1.0, c=None, coupling_term=None, run_t=1.0,
        int_dt=0.001):
    y = np.copy(start_y)
    yd = np.zeros(3)

    T = np.arange(0.0, run_t + dt, dt)
    Y = np.empty((len(T), len(y)))
    Y[0] = y
    for i in range(1, len(T)):
        spring_damper_step_quaternion(
            T[i - 1], T[i], y, yd, goal_y=goal_y, k=k, c=c,
            coupling_term=coupling_term, int_dt=int_dt)
        Y[i] = y
    return T, Y



import numpy as np
import pytransform3d.visualizer as pv
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
import pytransform3d.trajectories as ptr
from pytransform3d.batch_rotations import smooth_quaternion_trajectory
from mocap.cleaning import smooth_exponential_coordinates
from movement_primitives.kinematics import Kinematics
from movement_primitives.dmp import DualCartesianDMP
from movement_primitives.io import (write_json, write_yaml, write_pickle)
from movement_primitives.testing.simulation import get_absolute_path


def panel_pose(tm):
    tcp_left = tm.get_transform("LTCP_Link", "RH5_Root_Link")
    tcp_right = tm.get_transform("RTCP_Link", "RH5_Root_Link")
    tcp_left_pos = tcp_left[:3, 3]
    tcp_right_pos = tcp_right[:3, 3]
    tcp_middle = 0.5 * (tcp_left_pos + tcp_right_pos)
    x_axis = pr.norm_vector(tcp_right_pos - tcp_left_pos)
    y_axis = -pr.norm_vector(0.5 * (tcp_left[:3, 1] + tcp_right[:3, 1]))
    R_panel = pr.matrix_from_two_vectors(x_axis, y_axis)
    return pt.transform_from(R_panel, tcp_middle)


def animation_callback(step, graph, left_arm, right_arm, left_joint_trajectory, right_joint_trajectory, panel_mesh):
    left_arm.forward(left_joint_trajectory[step])
    right_arm.forward(right_joint_trajectory[step])
    graph.set_data()
    panel_mesh.set_data(panel_pose(graph.tm))
    return graph, panel_mesh


solar_panel_idx = 0
panel_rotation_angle = np.deg2rad(90)
n_steps = 201

with open(get_absolute_path("pybullet-only-arms-urdf/urdf/RH5.urdf", "models/robots/rh5_models"), "r") as f:
    kin = Kinematics(f.read(), mesh_path=get_absolute_path("pybullet-only-arms-urdf/urdf/", "models/robots/rh5_models"))
#kin.tm.write_png("graph.png", "twopi")
left_arm = kin.create_chain(
    ["ALShoulder1", "ALShoulder2", "ALShoulder3",
     "ALElbow", "ALWristRoll", "ALWristYaw", "ALWristPitch"],
    "RH5_Root_Link", "LTCP_Link")
right_arm = kin.create_chain(
    ["ARShoulder1", "ARShoulder2", "ARShoulder3",
     "ARElbow", "ARWristRoll", "ARWristYaw", "ARWristPitch"],
    "RH5_Root_Link", "RTCP_Link")

#q0_left = np.array([-1.57, 1.25, 0, -1.75, 0, 0, 0.8])
#q0_right = np.array([1.57, -1.25, 0, 1.75, 0, 0, 0.8])
q0_left = np.array([-1.57, 0.9, 0, -1.3, 0, 0, -0.55])
q0_right = np.array([1.57, -0.9, 0, 1.3, 0, 0, -0.55])

left_arm.forward(q0_left)
right_arm.forward(q0_right)

left2base_start = kin.tm.get_transform("LTCP_Link", "RH5_Root_Link")
right2base_start = kin.tm.get_transform("RTCP_Link", "RH5_Root_Link")
tcp_left_pos = left2base_start[:3, 3]
tcp_right_pos = right2base_start[:3, 3]

panel2base_start = panel_pose(kin.tm)
left2panel_start = pt.concat(left2base_start, pt.invert_transform(panel2base_start))
right2panel_start = pt.concat(right2base_start, pt.invert_transform(panel2base_start))


rotation_axis = -pr.unity

start2end = pt.rotate_transform(np.eye(4), pr.matrix_from_compact_axis_angle(rotation_axis * panel_rotation_angle))
left2panel_end = pt.concat(left2panel_start, start2end)
right2panel_end = pt.concat(right2panel_start, start2end)

left2base_end = pt.concat(left2panel_end, panel2base_start)
right2base_end = pt.concat(right2panel_end, panel2base_start)

start_left = pt.exponential_coordinates_from_transform(left2base_start)
end_left = pt.exponential_coordinates_from_transform(left2base_end)
start_right = pt.exponential_coordinates_from_transform(right2base_start)
end_right = pt.exponential_coordinates_from_transform(right2base_end)
start_left, end_left = smooth_exponential_coordinates(np.array([start_left, end_left]))
start_right, end_right = smooth_exponential_coordinates(np.array([start_right, end_right]))

T = np.linspace(0, 1, n_steps)
left_trajectory = start_left[np.newaxis] + T[:, np.newaxis] * (end_left[np.newaxis] - start_left[np.newaxis])
right_trajectory = start_right[np.newaxis] + T[:, np.newaxis] * (end_right[np.newaxis] - start_right[np.newaxis])

"""# axis-angle SLERP
import pytransform3d.batch_rotations as pbr
a_start = np.hstack((rotation_axis, (0,)))
a_end = np.hstack((rotation_axis, (panel_rotation_angle,)))
A = np.array([pr.axis_angle_slerp(a_start, a_end, t) for t in np.linspace(0, 1, n_steps)])
Rs = pbr.matrices_from_compact_axis_angles(A[:, :3] * A[:, 3, np.newaxis])
start2current = np.empty((n_steps, 4, 4))
start2current[:] = np.eye(4)
start2current[:, :3, :3] = Rs
left_trajectory = np.empty((n_steps, 4, 4))
right_trajectory = np.empty((n_steps, 4, 4))
for t in range(n_steps):
    left2panel_current = pt.concat(left2panel_start, start2current[t])
    left2base_current = pt.concat(left2panel_current, panel2base_start)
    left_trajectory[t] = left2base_current
    right2panel_current = pt.concat(right2panel_start, start2current[t])
    right2base_current = pt.concat(right2panel_current, panel2base_start)
    right_trajectory[t] = right2base_current
left_trajectory = ptr.exponential_coordinates_from_transforms(left_trajectory)
right_trajectory = ptr.exponential_coordinates_from_transforms(right_trajectory)
#"""

print("Imitation...")
dt = 0.01
execution_time = (n_steps - 1) * dt
T = np.linspace(0, execution_time, n_steps)
Y = np.hstack((smooth_exponential_coordinates(left_trajectory),
               smooth_exponential_coordinates(right_trajectory)))
#dmp = DMP(n_dims=Y.shape[1], execution_time=execution_time, dt=dt, n_weights_per_dim=10)
#dmp.imitate(T, Y)
#_, Y = dmp.open_loop()

########################################################################################################################
# DMP export

########################################################################################################################

left_trajectory = Y[:, :6]
right_trajectory = Y[:, 6:]

left_trajectory = ptr.transforms_from_exponential_coordinates(left_trajectory)
right_trajectory = ptr.transforms_from_exponential_coordinates(right_trajectory)

print("Inverse kinematics...")
random_state = np.random.RandomState(0)
left_joint_trajectory = left_arm.inverse_trajectory(left_trajectory, q0_left, random_state=random_state)
right_joint_trajectory = right_arm.inverse_trajectory(right_trajectory, q0_right, random_state=random_state)

########################################################################################################################
# Data export
lwp2ltcp = kin.tm.get_transform("ALWristPitch_Link", "LTCP_Link")
rwp2rtcp = kin.tm.get_transform("ARWristPitch_Link", "RTCP_Link")

lwp_trajectory = np.array([pt.concat(lwp2ltcp, ltcp2base) for ltcp2base in left_trajectory])
rwp_trajectory = np.array([pt.concat(rwp2rtcp, rtcp2base) for rtcp2base in right_trajectory])

left_trajectory_pq = ptr.pqs_from_transforms(lwp_trajectory)
right_trajectory_pq = ptr.pqs_from_transforms(rwp_trajectory)

dmp = DualCartesianDMP(execution_time=execution_time, dt=dt, n_weights_per_dim=10)
Y_pq = np.empty((len(Y), 14))
Y_pq[:, :7] = left_trajectory_pq
Y_pq[:, 7:] = right_trajectory_pq
Y_pq[:, 3:7] = smooth_quaternion_trajectory(Y_pq[:, 3:7])
Y_pq[:, 10:14] = smooth_quaternion_trajectory(Y_pq[:, 10:14])
dmp.imitate(T, Y_pq)

write_yaml("rh5_dual_arm_dmp.yaml", dmp)
write_json("rh5_dual_arm_dmp.json", dmp)
write_pickle("rh5_dual_arm_dmp.pickle", dmp)

import pandas as pd
data_export = np.hstack((
    left_trajectory_pq,
    right_trajectory_pq
))
df = pd.DataFrame(
    data=data_export, columns=[
    "left_pos_x", "left_pos_y", "left_pos_z", "left_ori_w", "left_ori_x", "left_ori_y", "left_ori_z",
    "right_pos_x", "right_pos_y", "right_pos_z", "right_ori_w", "right_ori_x", "right_ori_y", "right_ori_z"],
    index=pd.Index(np.linspace(0, execution_time, len(left_trajectory_pq)), name="Time"))
df.to_csv("rh5_dual_arm_rotate_panel_posquat.csv")

data_export = np.hstack((left_joint_trajectory, right_joint_trajectory))
df = pd.DataFrame(
    data=data_export, columns=left_arm.joint_names + right_arm.joint_names,
    index=pd.Index(np.linspace(0, execution_time, len(left_trajectory_pq)), name="Time"))
df.to_csv("rh5_dual_arm_rotate_panel_joints.csv")# float_format="%.20f"
########################################################################################################################

#"""
import matplotlib.pyplot as plt
from movement_primitives.plot import plot_trajectory_in_rows
axes = plot_trajectory_in_rows(np.hstack((left_joint_trajectory, right_joint_trajectory)))
for i, jn in enumerate(left_arm.joint_names + right_arm.joint_names):
    joint_limits = kin.tm._joints[jn][-2]
    axes[i].plot([0, len(left_joint_trajectory)], [joint_limits[0]] * 2, c="r")
    axes[i].plot([0, len(left_joint_trajectory)], [joint_limits[1]] * 2, c="r")
plt.show()
#"""

fig = pv.figure()
fig.plot_transform(s=0.3)

# "solar_panels/solar_panel_02/meshes/stl/base_link.stl"
# "solar_panels/solar_panel_03/meshes/stl/base link.stl"
panel_mesh = fig.plot_mesh("solar_panels/solar_panel_02/meshes/stl/base_link.stl", A2B=panel2base_start)

graph = fig.plot_graph(
    kin.tm, "RH5_Root_Link", show_visuals=True, show_collision_objects=False, show_frames=True, s=0.1,
    whitelist=["ALWristPitch_Link", "ARWristPitch_Link", "LTCP_Link", "RTCP_Link"])

fig.plot_transform(panel2base_start, s=0.2)

fig.plot_transform(left2base_start, s=0.15)
fig.plot_transform(right2base_start, s=0.15)
fig.plot_transform(left_trajectory[-1], s=0.15)
fig.plot_transform(right_trajectory[-1], s=0.15)

#pv.Trajectory(left_trajectory, s=0.05).add_artist(fig)
#pv.Trajectory(right_trajectory, s=0.05).add_artist(fig)

pv.Trajectory(lwp_trajectory, s=0.05).add_artist(fig)
pv.Trajectory(rwp_trajectory, s=0.05).add_artist(fig)

fig.view_init()
fig.animate(
    animation_callback, len(left_joint_trajectory), loop=True,
    fargs=(graph, left_arm, right_arm, left_joint_trajectory, right_joint_trajectory, panel_mesh))
fig.show()



"""Minimum jerk trajectory."""
import numpy as np
from .base import PointToPointMovement
from .data._minimum_jerk import generate_minimum_jerk


class MinimumJerkTrajectory(PointToPointMovement):
    """Precomputed point to point movement with minimum jerk.

    Parameters
    ----------
    n_dims : int
        State space dimensions.

    execution_time : float
        Execution time of the DMP.

    dt : float, optional (default: 0.01)
        Time difference between DMP steps.
    """
    def __init__(self, n_dims, execution_time, dt=0.01):
        super(MinimumJerkTrajectory, self).__init__(n_dims, n_dims)
        self.X = None
        self.Xd = None
        self.execution_time = execution_time
        self.dt = dt
        self.step_idx = 0
        self.initialized = False

    def reset(self):
        """Reset initial state and time."""
        self.step_idx = 0
        self.initialized = False

    def step(self, last_y, last_yd):
        """Perform step.

        Parameters
        ----------
        last_y : array, shape (n_dims,)
            Last state.

        last_yd : array, shape (n_dims,)
            Last time derivative of state (e.g., velocity).

        Returns
        -------
        y : array, shape (n_dims,)
            Next state.

        yd : array, shape (n_dims,)
            Next time derivative of state (e.g., velocity).
        """
        if not self.initialized:
            self.X, self.Xd, _ = generate_minimum_jerk(
                self.start_y, self.goal_y, self.execution_time, self.dt)
            self.initialized = True

        self.current_y = self.X[self.step_idx]
        self.current_yd = self.Xd[self.step_idx]
        self.step_idx += 1

        return np.copy(self.current_y), np.copy(self.current_yd)



"""
======================
Change DMP frequencies
======================

In this example, we modify the size of a time step in a DMP and compare its
effect.
"""
print(__doc__)


import matplotlib.pyplot as plt
import numpy as np
from movement_primitives.dmp import DMP

start_y = np.array([0.0])
goal_y = np.array([1.0])
execution_time = 1.0

dmp = DMP(n_dims=1, execution_time=execution_time)
dmp.configure(start_y=start_y, goal_y=goal_y)

plt.figure(figsize=(5, 8))
ax = plt.subplot(111)
ax.set_xlabel("Time")
ax.set_ylabel("Position")
for dt in [0.1, 0.075, 0.01]:
    # modify dt after DMP has been created
    dmp.dt_ = dt
    T, Y = dmp.open_loop(run_t=0.3)
    ax.scatter(T, Y, s=5)
    ax.plot(T, Y, label="dt = %g" % dt)
ax.legend()
plt.show()



"""
================
Critical Damping
================

The transformation system of a DMP converges to the goal and the convergence is
modeled as a spring-damper system. For an optimal convergence, the constants
defining the spring-damper system (spring constant k and damping coefficient c)
have to be set to critical damping for optimal convergence, that is, as quickly
as possible without overshooting. To illustrate this, we use a standalone
spring-damper system and explore several values for these parameters.
"""
print(__doc__)


import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.spring_damper import SpringDamper


k = 100
start_y = np.zeros(1)
goal_y = np.ones(1)
for c in [10, 20, 40]:
    attractor = SpringDamper(n_dims=1, k=k, c=c, dt=0.01)
    attractor.configure(start_y=start_y, goal_y=goal_y)
    T, Y = attractor.open_loop(run_t=2.0)
    plt.plot(T, Y[:, 0], label=f"$k={k}, c={c}$")
plt.scatter(1.0, 1.0, marker="*", s=200, label="Goal")
plt.legend(loc="best")
plt.title(r"Condition for critical damping: $c = 2 \sqrt{k}$")
plt.show()



"""
===================
State-following DMP
===================
"""
print(__doc__)


import matplotlib.pyplot as plt
import numpy as np
from movement_primitives.dmp import StateFollowingDMP

start_y = np.array([0.0, 1.0])
dt = 0.001
execution_time = 3.0
n_viapoints = 8

dmp = StateFollowingDMP(n_dims=2, execution_time=execution_time, dt=dt, n_viapoints=n_viapoints)
dmp.forcing_term.viapoints[:, 0] = np.linspace(0, 1, n_viapoints)
dmp.forcing_term.viapoints[:, 1] = np.linspace(1, 2, n_viapoints)
dmp.configure(start_y=start_y)

T, Y = dmp.open_loop(run_t=1.5 * execution_time)

plt.figure(figsize=(10, 5))
ax = plt.subplot(121)
ax.set_xlabel("Time")
ax.set_ylabel("Position")
ax.plot(T, Y[:, 0], label="X")
ax.plot(T, Y[:, 1], label="Y")
ax.scatter(np.linspace(0, execution_time, n_viapoints), dmp.forcing_term.viapoints[:, 0], label="Viapoints X")
ax.scatter(np.linspace(0, execution_time, n_viapoints), dmp.forcing_term.viapoints[:, 1], label="Viapoints Y")
ax.legend()
ax = plt.subplot(122)
ax.set_xlabel("Position X")
ax.set_ylabel("Position Y")
ax.plot(Y[:, 0], Y[:, 1], label="Positions")
ax.scatter(dmp.forcing_term.viapoints[:, 0], dmp.forcing_term.viapoints[:, 1], label="Viapoints")
ax.legend()
plt.show()



from functools import partial
import numpy as np
from movement_primitives.dmp._canonical_system import canonical_system_alpha
from movement_primitives.dmp._canonical_system import phase as phase_python
from movement_primitives.dmp_fast import phase as phase_cython
import timeit


goal_t = 1.0
start_t = 0.0
alpha = canonical_system_alpha(0.01, goal_t, start_t)
times = timeit.repeat(partial(phase_python, 0.5, alpha, goal_t, start_t), repeat=1000, number=1000)
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))
times = timeit.repeat(partial(phase_cython, 0.5, alpha, goal_t, start_t), repeat=1000, number=1000)
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))



import numpy as np
import pytransform3d.visualizer as pv
from pytransform3d.urdf import UrdfTransformManager
import pytransform3d.transformations as pt
from gmr import GMM
from mocap.dataset_loader import load_kuka_dataset, transpose_dataset, smooth_dual_arm_trajectories_pq

from movement_primitives.promp import ProMP
from movement_primitives.visualization import to_ellipsoid
from movement_primitives.kinematics import Kinematics


def generate_training_data(
        pattern, n_weights_per_dim, context_names, smooth_quaterions, verbose=0):
    Ts, Ps, contexts = transpose_dataset(
        load_kuka_dataset(pattern, context_names, verbose=verbose))

    if smooth_quaterions:
        smooth_dual_arm_trajectories_pq(Ps)

    n_demos = len(Ts)
    n_dims = Ps[0].shape[1]

    promp = ProMP(n_dims=n_dims, n_weights_per_dim=n_weights_per_dim)
    weights = np.empty((n_demos, n_dims * n_weights_per_dim))
    for demo_idx in range(n_demos):
        weights[demo_idx] = promp.weights(Ts[demo_idx], Ps[demo_idx])

    return weights, Ts, Ps, contexts


plot_training_data = False
n_dims = 14
n_weights_per_dim = 10
# available contexts: "panel_width", "clockwise", "counterclockwise", "left_arm", "right_arm"
context_names = ["panel_width", "clockwise", "counterclockwise"]

#pattern = "data/kuka/20200129_peg_in_hole/csv_processed/*/*.csv"
#pattern = "data/kuka/20191213_carry_heavy_load/csv_processed/*/*.csv"
pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/*/*.csv"

weights, Ts, Ps, contexts = generate_training_data(
    pattern, n_weights_per_dim, context_names=context_names,
    smooth_quaterions=True, verbose=2)
X = np.hstack((contexts, weights))

random_state = np.random.RandomState(0)
gmm = GMM(n_components=5, random_state=random_state)
gmm.from_samples(X)

with open("kuka_lbr/urdf/kuka_lbr.urdf", "r") as f:
    kin = Kinematics(f.read(), mesh_path="kuka_lbr/urdf/")
right_chain = kin.create_chain(
    ["kuka_lbr_r_joint_%d" % i for i in range(1, 8)],
    "kuka_lbr", "kuka_lbr_r_tcp", verbose=0)
left_chain = kin.create_chain(
    ["kuka_lbr_l_joint_%d" % i for i in range(1, 8)],
    "kuka_lbr", "kuka_lbr_l_tcp", verbose=0)
mean_start = np.mean([
    Ps[i][0] for i in range(len(Ps)) if contexts[i][0] == 0.5], axis=0)
mean_start_left = pt.transform_from_pq(mean_start[:7])
mean_start_right = pt.transform_from_pq(mean_start[7:])
left_chain.inverse(mean_start_left, initial_joint_angles=np.zeros(7))
right_chain.inverse(mean_start_right, initial_joint_angles=np.zeros(7))

n_validation_samples = 100
n_steps = 100
T_query = np.linspace(0, 1, n_steps)

fig = pv.figure(with_key_callbacks=True)
fig.plot_transform(s=0.1)
tm = UrdfTransformManager()
fig.plot_graph(
    kin.tm, "kuka_lbr", show_frames=False, show_visuals=True,
    whitelist=["kuka_lbr_l_tcp", "kuka_lbr_r_tcp"], s=0.2)

for panel_width, color, idx in zip([0.3, 0.4, 0.5], ([1.0, 1.0, 0.0], [0.0, 1.0, 1.0], [1.0, 0.0, 1.0]), range(3)):
    print("panel_width = %.2f, color = %s" % (panel_width, color))

    context = np.array([panel_width, 0.0, 1.0])
    conditional_weight_distribution = gmm.condition(np.arange(len(context)), context).to_mvn()
    promp = ProMP(n_dims=n_dims, n_weights_per_dim=n_weights_per_dim)
    promp.from_weight_distribution(
        conditional_weight_distribution.mean,
        conditional_weight_distribution.covariance)

    # mean and covariance in state space
    mean = promp.mean_trajectory(T_query)
    cov = promp.cov_trajectory(T_query).reshape(
        mean.shape[0], mean.shape[1], mean.shape[1], mean.shape[0])

    for t in range(0, len(mean), 1):
        p_left = mean[t, :3]
        p_right = mean[t, 7:10]
        cov_left = cov[t, :3, :3, t]
        cov_right = cov[t, 7:10, 7:10, t]
        ellipsoid2origin, radii = to_ellipsoid(p_left, cov_left)
        fig.plot_ellipsoid(A2B=ellipsoid2origin, radii=0.5 * radii, c=color)
        ellipsoid2origin, radii = to_ellipsoid(p_right, cov_right)
        fig.plot_ellipsoid(A2B=ellipsoid2origin, radii=0.5 * radii, c=color)

if plot_training_data:
    for P in Ps:
        left = fig.plot_trajectory(P=P[:, :7], s=0.02)
        right = fig.plot_trajectory(P=P[:, 7:], s=0.02)

fig.view_init(azim=0, elev=25)
fig.show()



import numpy as np


def generate_1d_trajectory_distribution(
        n_demos, n_steps, initial_offset_range=3.0, final_offset_range=0.1,
        noise_per_step_range=20.0, random_state=np.random.RandomState(0)):
    """Generates toy data for testing and demonstration.

    Parameters
    ----------
    n_demos : int
        Number of demonstrations

    n_steps : int
        Number of steps

    initial_offset_range : float, optional (default: 3)
        Range of initial offset from cosine

    final_offset_range : float, optional (default: 0.1)
        Range of final offset from cosine

    noise_per_step_range : float, optional (default: 20)
        Factor for noise in each step

    random_state : RandomState, optional (default: seed 0)
        Random state

    Returns
    -------
    T : array, shape (n_steps,)
        Times

    Y : array, shape (n_demos, n_steps, 1)
        Demonstrations (positions)
    """
    T = np.linspace(0, 1, n_steps)
    Y = np.empty((n_demos, n_steps, 1))

    A = create_finite_differences_matrix_1d(n_steps, dt=1.0 / (n_steps - 1))
    cov = np.linalg.inv(A.T.dot(A))
    L = np.linalg.cholesky(cov)

    for demo_idx in range(n_demos):
        Y[demo_idx, :, 0] = np.cos(2 * np.pi * T)
        if initial_offset_range or final_offset_range:
            initial_offset = initial_offset_range * (random_state.rand() - 0.5)
            final_offset = final_offset_range * (random_state.rand() - 0.5)
            Y[demo_idx, :, 0] += np.linspace(
                initial_offset, final_offset, n_steps)
        if noise_per_step_range:
            noise_per_step = (noise_per_step_range
                              * L.dot(random_state.randn(n_steps)))
            Y[demo_idx, :, 0] += noise_per_step
    return T, Y


def create_finite_differences_matrix_1d(n_steps, dt):
    """Finite difference matrix to compute accelerations from positions.

    Parameters
    ----------
    n_steps : int
        Number of steps of the resulting trajectory.

    dt : float
        Time between steps.

    Returns
    -------
    A : array, shape (n_steps + 2, n_steps)
        Finite difference matrix.
    """
    A = np.zeros((n_steps + 2, n_steps), dtype=float)
    super_diagonal = (np.arange(n_steps), np.arange(n_steps))
    sub_diagonal = (np.arange(2, n_steps + 2), np.arange(n_steps))
    A[super_diagonal] = np.ones(n_steps)
    A[sub_diagonal] = np.ones(n_steps)
    main_diagonal = (np.arange(1, n_steps + 1), np.arange(n_steps))
    A[main_diagonal] = -2 * np.ones(n_steps)
    return A / (dt ** 2)



"""
=================
Conditional ProMP
=================

Probabilistic Movement Primitives (ProMPs) define distributions over
trajectories that can be conditioned on viapoints. In this example, we
plot the resulting posterior distribution after conditioning on varying
start positions.
"""
print(__doc__)


import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.data import generate_1d_trajectory_distribution
from movement_primitives.promp import ProMP


n_demos = 100
n_steps = 101
T, Y = generate_1d_trajectory_distribution(n_demos, n_steps)
y_conditional_cov = np.array([0.025])
promp = ProMP(n_dims=1, n_weights_per_dim=10)
promp.imitate([T] * n_demos, Y)
Y_mean = promp.mean_trajectory(T)
Y_conf = 1.96 * np.sqrt(promp.var_trajectory(T))

plt.figure(figsize=(10, 5))

ax1 = plt.subplot(121)
ax1.set_title("Training set and ProMP")
ax1.fill_between(T, (Y_mean - Y_conf).ravel(), (Y_mean + Y_conf).ravel(), color="r", alpha=0.3)
ax1.plot(T, Y_mean, c="r", lw=2, label="ProMP")
ax1.plot(T, Y[:, :, 0].T, c="k", alpha=0.1)
ax1.set_xlim((-0.05, 1.05))
ax1.set_ylim((-2.5, 3))
ax1.legend(loc="best")

ax2 = plt.subplot(122)
ax2.set_title("Conditional ProMPs")

for color, y_cond in zip("rgbcmyk", np.linspace(-1, 2.5, 7)):
    cpromp = promp.condition_position(np.array([y_cond]), y_cov=y_conditional_cov, t=0.0, t_max=1.0)
    Y_cmean = cpromp.mean_trajectory(T)
    Y_cconf = 1.96 * np.sqrt(cpromp.var_trajectory(T))

    ax2.scatter([0], [y_cond], marker="*", s=100, c=color, label="$y_0 = %.2f$" % y_cond)
    ax2.fill_between(T, (Y_cmean - Y_cconf).ravel(), (Y_cmean + Y_cconf).ravel(), color=color, alpha=0.3)
    ax2.plot(T, Y_cmean, c=color, lw=2)
    ax2.set_xlim((-0.05, 1.05))
    ax2.set_ylim((-2.5, 3))
    ax2.legend(loc="best")

ax1.set_xlabel("Time $t$ [s]")
ax1.set_ylabel("Position $y$ [m]")
ax2.set_xlabel("Time $t$ [s]")
plt.tight_layout()
plt.show()



"""Robot simulations based on PyBullet.

Note that PyBullet internally represents orientations as quaternions with
scalar-last convention: (qx, qy, qz, qw).
"""
import numpy as np
import os
try:
    import pybullet
    import pybullet_data
    pybullet_available = True
except ImportError:
    pybullet_available = False
import pytransform3d.transformations as pt


class PybulletSimulation:
    """PyBullet simulation of a robot.

    Parameters
    ----------
    dt : float
        Length of time step

    gui : bool, optional (default: True)
        Show PyBullet GUI

    real_time : bool, optional (default: False)
        Simulate in real time
    """
    def __init__(self, dt, gui=True, real_time=False):
        assert pybullet_available
        self.dt = dt
        if gui:
            self.client_id = pybullet.connect(pybullet.GUI)
        else:
            self.client_id = pybullet.connect(pybullet.DIRECT)

        pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_GUI, 0)
        pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_SHADOWS, 0)
        pybullet.resetDebugVisualizerCamera(2, 75, -30, [0, 0, 0])

        pybullet.resetSimulation(physicsClientId=self.client_id)
        pybullet.setTimeStep(dt, physicsClientId=self.client_id)
        pybullet.setRealTimeSimulation(
            1 if real_time else 0, physicsClientId=self.client_id)
        pybullet.setGravity(0, 0, -9.81, physicsClientId=self.client_id)

    def step(self):
        """Simulation step."""
        assert pybullet.isConnected(self.client_id)
        pybullet.stepSimulation(physicsClientId=self.client_id)

    def sim_loop(self, n_steps=None):
        """Run simulation loop.

        Parameters
        ----------
        n_steps : int, optional (default: infinite)
            Number of simulation steps.
        """
        if n_steps is None:
            while pybullet.isConnected(self.client_id):
                pybullet.stepSimulation(physicsClientId=self.client_id)
        else:
            for _ in range(n_steps):
                if not pybullet.isConnected(self.client_id):
                    break
                pybullet.stepSimulation(physicsClientId=self.client_id)


def _pybullet_pose(pose):
    """Convert pose from (x, y, z, qw, qx, qy, qz) to ((x, y, z), (qx, qy, qz, qw))."""
    pos = pose[:3]
    rot = pose[3:]
    rot = np.hstack((rot[1:], [rot[0]]))  # wxyz -> xyzw
    return pos, rot


def _pytransform_pose(pos, rot):
    """Convert pose from ((x, y, z), (qx, qy, qz, qw)) to (x, y, z, qw, qx, qy, qz)."""
    return np.hstack((pos, [rot[-1]], rot[:-1]))  # xyzw -> wxyz


def draw_transform(pose2origin, s, client_id, lw=1):
    """Draw pose represented by transformation matrix.

    Parameters
    ----------
    pose2origin : array-like, shape (4, 4)
        Homogeneous transformation matrix

    s : float
        Scale, length of the coordinate axes

    client_id : int
        Physics client ID

    lw : int, optional (default: 1)
        Line width
    """
    pose2origin = pt.check_transform(pose2origin)
    pybullet.addUserDebugLine(
        pose2origin[:3, 3], pose2origin[:3, 3] + s * pose2origin[:3, 0],
        [1, 0, 0], lw, physicsClientId=client_id)
    pybullet.addUserDebugLine(
        pose2origin[:3, 3], pose2origin[:3, 3] + s * pose2origin[:3, 1],
        [0, 1, 0], lw, physicsClientId=client_id)
    pybullet.addUserDebugLine(
        pose2origin[:3, 3], pose2origin[:3, 3] + s * pose2origin[:3, 2],
        [0, 0, 1], lw, physicsClientId=client_id)


def draw_pose(pose2origin, s, client_id, lw=1):
    """Draw pose represented by position and quaternion.

    Parameters
    ----------
    pose2origin : array-like, shape (7,)
        Position and quaternion: (x, y, z, qw, qx, qy, qz)

    s : float
        Scale, length of the coordinate axes

    client_id : int
        Physics client ID

    lw : int, optional (default: 1)
        Line width
    """
    pose2origin = pt.transform_from_pq(pose2origin)
    pybullet.addUserDebugLine(
        pose2origin[:3, 3], pose2origin[:3, 3] + s * pose2origin[:3, 0],
        [1, 0, 0], lw, physicsClientId=client_id)
    pybullet.addUserDebugLine(
        pose2origin[:3, 3], pose2origin[:3, 3] + s * pose2origin[:3, 1],
        [0, 1, 0], lw, physicsClientId=client_id)
    pybullet.addUserDebugLine(
        pose2origin[:3, 3], pose2origin[:3, 3] + s * pose2origin[:3, 2],
        [0, 0, 1], lw, physicsClientId=client_id)


def draw_trajectory(A2Bs, client_id, n_key_frames=10, s=1.0, lw=1):
    """Draw trajectory.

    Parameters
    ----------
    A2Bs : array-like, shape (n_steps, 4, 4)
        Homogeneous transformation matrices

    client_id : int
        Physics client ID

    n_key_frames : int, optional (default: 10)
        Number of coordinate frames

    s : float, optional (default: 1)
        Scale, length of the coordinate axes

    lw : int, optional (default: 1)
        Line width
    """
    key_frames_indices = np.linspace(
        0, len(A2Bs) - 1, n_key_frames, dtype=np.int64)
    for idx in key_frames_indices:
        draw_transform(A2Bs[idx], s=s, client_id=client_id)
    for idx in range(len(A2Bs) - 1):
        pybullet.addUserDebugLine(
            A2Bs[idx, :3, 3], A2Bs[idx + 1, :3, 3], [0, 0, 0], lw,
            physicsClientId=client_id)


def get_absolute_path(urdf_path, model_prefix_path):
    autoproj_dir = None
    if "AUTOPROJ_CURRENT_ROOT" in os.environ and os.path.exists(os.environ["AUTOPROJ_CURRENT_ROOT"]):
        autoproj_dir = os.environ["AUTOPROJ_CURRENT_ROOT"]
    if autoproj_dir is not None and os.path.exists(os.path.join(autoproj_dir, model_prefix_path)):
        return os.path.join(autoproj_dir, model_prefix_path, urdf_path)
    else:
        return urdf_path


class UR5Simulation(PybulletSimulation):
    """PyBullet simulation of UR5 robot arm.

    Parameters
    ----------
    dt : float
        Length of time step

    gui : bool, optional (default: True)
        Show PyBullet GUI

    real_time : bool, optional (default: False)
        Simulate in real time
    """
    def __init__(self, dt, gui=True, real_time=False):
        super(UR5Simulation, self).__init__(dt, gui, real_time)

        pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())
        self.plane = pybullet.loadURDF(
            "plane.urdf", [0, 0, -1], useFixedBase=1)
        self.robot = pybullet.loadURDF(
            "examples/data/urdf/ur5.urdf", [0, 0, 0], useFixedBase=1)

        self.base_pose = pybullet.getBasePositionAndOrientation(self.robot)
        self.inv_base_pose = pybullet.invertTransform(*self.base_pose)

        self.n_ur5_joints = 6
        # one link after the base link of the last joint
        self.ee_link_index = pybullet.getJointInfo(
            self.robot, self.n_ur5_joints)[16] + 2

        self.n_joints = pybullet.getNumJoints(self.robot)
        self.joint_indices = [
            i for i in range(self.n_joints)
            if pybullet.getJointInfo(self.robot, i)[2] == 0]  # joint type 0: revolute
        self.joint_names = {i: pybullet.getJointInfo(self.robot, i)[1]
                            for i in self.joint_indices}
        # we cannot actually use them so far:
        self.joint_max_velocities = [pybullet.getJointInfo(self.robot, i)[11]
                                     for i in self.joint_indices]

    def inverse_kinematics(self, ee2robot):
        """Inverse kinematics of UR5.

        Parameters
        ----------
        ee2robot : array-like, shape (7,)
            End-effector pose: (x, y, z, qw, qx, qy, qz).

        Returns
        -------
        q : array, shape (n_joints,)
            Joint angles
        """
        pos, rot = _pybullet_pose(ee2robot)
        # ee2world
        pos, rot = pybullet.multiplyTransforms(pos, rot, *self.base_pose)

        q = pybullet.calculateInverseKinematics(
            self.robot, self.ee_link_index, pos, rot, maxNumIterations=100,
            residualThreshold=0.001)
        q = q[:self.n_ur5_joints]
        if any(np.isnan(q)):
            raise Exception("IK solver found no solution.")
        return q

    def get_joint_state(self):
        """Get joint state.

        Returns
        -------
        positions : array, shape (n_joints,)
            Joint angles

        velocities : array, shape (n_joints,)
            Joint velocities
        """
        joint_states = pybullet.getJointStates(self.robot, self.joint_indices[:self.n_ur5_joints])
        positions = []
        velocities = []
        for joint_state in joint_states:
            pos, vel, forces, torque = joint_state
            positions.append(pos)
            velocities.append(vel)
        return np.asarray(positions), np.asarray(velocities)

    def set_desired_joint_state(self, joint_state, position_control=False):
        if position_control:
            pybullet.setJointMotorControlArray(
                self.robot, self.joint_indices[:self.n_ur5_joints],
                pybullet.POSITION_CONTROL,
                targetPositions=joint_state)
        else:  # velocity control
            pybullet.setJointMotorControlArray(
                self.robot, self.joint_indices[:self.n_ur5_joints],
                pybullet.VELOCITY_CONTROL, targetVelocities=joint_state)

    def get_ee_state(self, return_velocity=False):
        ee_state = pybullet.getLinkState(
            self.robot, self.ee_link_index, computeLinkVelocity=1,
            computeForwardKinematics=1)
        pos = ee_state[4]
        rot = ee_state[5]
        pos, rot = pybullet.multiplyTransforms(pos, rot, *self.inv_base_pose)
        if return_velocity:
            vel = ee_state[6]
            #ang_vel = ee_state[7]
            #ang_speed = np.linalg.norm(ang_vel)
            #ang_axis = np.asarray(ang_vel) / ang_speed
            vel, _ = pybullet.multiplyTransforms(
                vel, [0, 0, 0, 1], *self.inv_base_pose)
            # TODO transform angular velocity?
            return _pytransform_pose(pos, rot), np.hstack((vel, np.zeros(3)))
        else:
            return _pytransform_pose(pos, rot)

    def set_desired_ee_state(self, ee_state):
        q = self.inverse_kinematics(ee_state)
        last_q, last_qd = self.get_joint_state()
        self.set_desired_joint_state(
            (q - last_q) / self.dt, position_control=False)

    def stop(self):
        pybullet.setJointMotorControlArray(
            self.robot, self.joint_indices[:self.n_ur5_joints],
            pybullet.VELOCITY_CONTROL,
            targetVelocities=np.zeros(self.n_ur5_joints))
        self.step()

    def goto_ee_state(self, ee_state, wait_time=1.0, text=None):
        if text:
            pos, rot = _pybullet_pose(ee_state)
            self.write(pos, text)
        q = self.inverse_kinematics(ee_state)
        self.set_desired_joint_state(q, position_control=True)
        self.sim_loop(int(wait_time / self.dt))

    def step_through_cartesian(self, steppable, last_p, last_v, execution_time, closed_loop=False):
        p, v = self.get_ee_state(return_velocity=True)
        desired_positions = [last_p]
        positions = [p]
        desired_velocities = [last_v]
        velocities = [v]

        for i in range(int(execution_time / self.dt)):
            if closed_loop:
                last_p, _ = self.get_ee_state(return_velocity=True)  # TODO last_v

            p, v = steppable.step(last_p, last_v)
            self.set_desired_ee_state(p)
            self.step()

            desired_positions.append(p)
            desired_velocities.append(v)

            last_v = v
            last_p = p

            p, v = self.get_ee_state(return_velocity=True)
            positions.append(p)
            velocities.append(v)

        self.stop()

        return (np.asarray(desired_positions),
                np.asarray(positions),
                np.asarray(desired_velocities),
                np.asarray(velocities))

    def write(self, pos, text):
        pybullet.addUserDebugText(text, pos, [0, 0, 0])
        pybullet.addUserDebugLine(pos, [0, 0, 0], [0, 0, 0], 2)


class KinematicsChain:
    """Wrapper of PyBullet simulation of one kinematic chain.

    This can be used to compute inverse kinematics.

    Parameters
    ----------
    ee_frame : str
        Frame of the end-effector.

    joints : list
        Names of joints.

    urdf_path : str
        Path to URDF file.

    debug_gui : bool, optional (default: False)
        Start PyBullet with GUI for debugging purposes.

    Attributes
    ----------
    client_id_ : int
        ID of the PyBullet instance.
    """
    def __init__(self, ee_frame, joints, urdf_path, debug_gui=False):
        if debug_gui:
            self.client_id_ = pybullet.connect(pybullet.GUI)
        else:
            self.client_id_ = pybullet.connect(pybullet.DIRECT)
        pybullet.resetSimulation(physicsClientId=self.client_id_)
        pybullet.setTimeStep(1.0, physicsClientId=self.client_id_)

        self.chain = pybullet.loadURDF(
            urdf_path, useFixedBase=1, physicsClientId=self.client_id_)
        self.joint_indices, self.link_indices, \
        self.joint_name_to_joint_idx_ik = analyze_robot(
            robot=self.chain, physicsClientId=self.client_id_,
            return_joint_indices=True, verbose=0)

        self.chain_joint_indices = [self.joint_indices[jn] for jn in joints]
        self.ee_idx = self.link_indices[ee_frame]
        self.q_indices = [self.joint_name_to_joint_idx_ik[jn] for jn in joints]

        self.n_all_chain_joints = pybullet.getNumJoints(
            self.chain, self.client_id_)

    def inverse(self, desired_ee_state, q_current=None, n_iter=100,
                threshold=0.001):
        """Compute inverse kinematics.

        Parameters
        ----------
        desired_ee_state : array-like, shape (7,)
            Desired pose of end effector: (x, y, z, qw, qx, qy, qz)

        q_current : array-like, shape (n_joints,)
            Current joint angles.

        n_iter : int, optional (default: 100)
            Number of iterations for the numerical inverse kinematics solver.

        threshold : float, optional (default: 0.001)
            Threshold for residual.

        Returns
        -------
        q : array, shape (n_joints,)
            Joint angles to reach the desired pose of the end effector.
            If the desired pose cannot be reached, this function will return
            joint angles that come as close as possible.
        """
        if q_current is not None:
            # we have to actively go to the current joint state
            # before computing inverse kinematics
            self.goto_joint_state(q_current)

        ee_pos, ee_rot = _pybullet_pose(desired_ee_state)
        q = pybullet.calculateInverseKinematics(
            self.chain, self.ee_idx, ee_pos, ee_rot,
            maxNumIterations=n_iter, residualThreshold=threshold,
            jointDamping=[0.1] * self.n_all_chain_joints,
            physicsClientId=self.client_id_)
        return np.asarray(q)[self.q_indices]

    def goto_joint_state(self, q_current, max_steps_to_joint_state=50,
                         joint_state_eps=0.001):
        pybullet.setJointMotorControlArray(
            self.chain, self.chain_joint_indices,
            pybullet.POSITION_CONTROL,
            targetPositions=q_current, physicsClientId=self.client_id_)

        for _ in range(max_steps_to_joint_state):
            pybullet.stepSimulation(physicsClientId=self.client_id_)
            q_internal = np.array([js[0] for js in pybullet.getJointStates(
                self.chain, self.chain_joint_indices,
                physicsClientId=self.client_id_)])
            if np.linalg.norm(q_current - q_internal) < joint_state_eps:
                break


class RH5Simulation(PybulletSimulation):
    """PyBullet simulation of RH5 humanoid robot.

    Parameters
    ----------
    dt : float
        Length of time step

    gui : bool, optional (default: True)
        Show PyBullet GUI

    real_time : bool, optional (default: False)
        Simulate in real time

    left_joints : tuple, optional
        Joints of the left arm

    right_joints : tuple, optional
        Joints of the right arm

    urdf_path : str, optional
        Path to URDF file

    left_arm_path : str, optional
        Path to URDF of left arm

    right_arm_path : str, optional
        Path to URDF of right arm
    """
    def __init__(self, dt, gui=True, real_time=False,
                 left_ee_frame="LTCP_Link", right_ee_frame="RTCP_Link",
                 left_joints=("ALShoulder1", "ALShoulder2", "ALShoulder3", "ALElbow", "ALWristRoll", "ALWristYaw", "ALWristPitch"),
                 right_joints=("ARShoulder1", "ARShoulder2", "ARShoulder3", "ARElbow", "ARWristRoll", "ARWristYaw", "ARWristPitch"),
                 urdf_path=get_absolute_path("pybullet-only-arms-urdf/urdf/RH5.urdf", "models/robots/rh5_models"),
                 left_arm_path=get_absolute_path("pybullet-only-arms-urdf/submodels/left_arm.urdf", "models/robots/rh5_models"),
                 right_arm_path=get_absolute_path("pybullet-only-arms-urdf/submodels/right_arm.urdf", "models/robots/rh5_models")):
        super(RH5Simulation, self).__init__(dt, gui, real_time)

        self.base_pos = (0, 0, 0)

        pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())
        self.plane = pybullet.loadURDF(
            "plane.urdf", (0, 0, -1), useFixedBase=1,
            physicsClientId=self.client_id)
        self.robot = pybullet.loadURDF(
            urdf_path, self.base_pos, useFixedBase=1,
            physicsClientId=self.client_id)
        self.joint_indices, self.link_indices = analyze_robot(
            robot=self.robot, physicsClientId=self.client_id)

        self.base_pose = self.base_pos, (0.0, 0.0, 0.0, 1.0)  # not pybullet.getBasePositionAndOrientation(self.robot)
        self.inv_base_pose = pybullet.invertTransform(*self.base_pose)

        self.n_joints = len(left_joints) + len(right_joints)
        self.n_left_joints = len(left_joints)
        self.left_arm_joint_indices = [self.joint_indices[jn] for jn in left_joints]
        self.right_arm_joint_indices = [self.joint_indices[jn] for jn in right_joints]
        self.left_ee_link_index = self.link_indices[left_ee_frame]
        self.right_ee_link_index = self.link_indices[right_ee_frame]

        self.left_ik = KinematicsChain(
            left_ee_frame, left_joints, left_arm_path)
        self.right_ik = KinematicsChain(
            right_ee_frame, right_joints, right_arm_path)

    def inverse_kinematics(self, ee2robot):
        """Inverse kinematics of RH5.

        Parameters
        ----------
        ee2robot : array-like, shape (14,)
            End-effector poses: two 7-tuples that contain the pose of the
            left and right end effector in the order (x, y, z, qw, qx, qy, qz).

        Returns
        -------
        q : array, shape (n_joints,)
            Joint angles
        """
        q = np.empty(self.n_joints)

        left_q = np.array([js[0] for js in pybullet.getJointStates(
            self.robot, self.left_arm_joint_indices, physicsClientId=self.client_id)])
        q[:self.n_left_joints] = self.left_ik.inverse(ee2robot[:7], left_q)

        right_q = np.array([js[0] for js in pybullet.getJointStates(
            self.robot, self.right_arm_joint_indices, physicsClientId=self.client_id)])
        q[self.n_left_joints:] = self.right_ik.inverse(ee2robot[7:], right_q)

        return q

    def get_joint_state(self):
        """Get joint state.

        Returns
        -------
        positions : array, shape (n_joints,)
            Joint angles

        velocities : array, shape (n_joints,)
            Joint velocities
        """
        joint_states = pybullet.getJointStates(
            self.robot, self.left_arm_joint_indices + self.right_arm_joint_indices,
            physicsClientId=self.client_id)
        positions = np.empty(self.n_joints)
        velocities = np.empty(self.n_joints)
        for joint_idx, joint_state in enumerate(joint_states):
            positions[joint_idx], velocities[joint_idx], forces, torque = joint_state
        return positions, velocities

    def set_desired_joint_state(self, joint_state, position_control=False):
        left_joint_state, right_joint_state = np.split(joint_state, (len(self.left_arm_joint_indices),))
        if position_control:
            pybullet.setJointMotorControlArray(
                self.robot, self.left_arm_joint_indices,
                pybullet.POSITION_CONTROL,
                targetPositions=left_joint_state,
                physicsClientId=self.client_id)
            pybullet.setJointMotorControlArray(
                self.robot, self.right_arm_joint_indices,
                pybullet.POSITION_CONTROL,
                targetPositions=right_joint_state,
                physicsClientId=self.client_id)
        else:  # velocity control
            pybullet.setJointMotorControlArray(
                self.robot, self.left_arm_joint_indices,
                pybullet.VELOCITY_CONTROL, targetVelocities=left_joint_state,
                physicsClientId=self.client_id)
            pybullet.setJointMotorControlArray(
                self.robot, self.right_arm_joint_indices,
                pybullet.VELOCITY_CONTROL, targetVelocities=right_joint_state,
                physicsClientId=self.client_id)

    def get_ee_state(self, return_velocity=False):
        left_ee_state = pybullet.getLinkState(
            self.robot, self.left_ee_link_index, computeLinkVelocity=1,
            computeForwardKinematics=1, physicsClientId=self.client_id)
        left_pos = left_ee_state[4]
        left_rot = left_ee_state[5]
        left_pos, left_rot = pybullet.multiplyTransforms(left_pos, left_rot, *self.inv_base_pose)
        left_pose = _pytransform_pose(left_pos, left_rot)

        right_ee_state = pybullet.getLinkState(
            self.robot, self.right_ee_link_index, computeLinkVelocity=1,
            computeForwardKinematics=1, physicsClientId=self.client_id)
        right_pos = right_ee_state[4]
        right_rot = right_ee_state[5]
        right_pos, right_rot = pybullet.multiplyTransforms(right_pos, right_rot, *self.inv_base_pose)
        right_pose = _pytransform_pose(right_pos, right_rot)

        if return_velocity:
            raise NotImplementedError()
            """
            left_vel = left_ee_state[6]
            #ang_vel = ee_state[7]
            #ang_speed = np.linalg.norm(ang_vel)
            #ang_axis = np.asarray(ang_vel) / ang_speed
            left_vel, _ = pybullet.multiplyTransforms(
                left_vel, [0, 0, 0, 1], *self.inv_base_pose)
            # TODO transform angular velocity?
            return _pytransform_pose(pos, rot), np.hstack((vel, np.zeros(3)))
            """
        else:
            return np.hstack((left_pose, right_pose))

    def set_desired_ee_state(self, ee_state, position_control=False):
        q = self.inverse_kinematics(ee_state)
        if position_control:
            self.set_desired_joint_state(q, position_control=True)
        else:
            last_q, _ = self.get_joint_state()
            self.set_desired_joint_state(
                (q - last_q) / self.dt, position_control=False)

    def stop(self):
        ee_state = self.get_ee_state(return_velocity=False)
        self.goto_ee_state(ee_state)
        self.step()

    def goto_ee_state(self, ee_state, wait_time=1.0, text=None):
        if text:
            pos, rot = _pybullet_pose(ee_state)
            self.write(pos, text)
        q = self.inverse_kinematics(ee_state)
        self.set_desired_joint_state(q, position_control=True)
        self.sim_loop(int(wait_time / self.dt))

    def step_through_cartesian(self, steppable, last_p, last_v, execution_time, closed_loop=False, coupling_term=None):
        p = self.get_ee_state(return_velocity=False)   # TODO v
        desired_positions = [last_p]
        positions = [p]
        desired_velocities = [last_v]
        velocities = [np.zeros(12)]

        for i in range(int(execution_time / self.dt)):
            if closed_loop:
                last_p = self.get_ee_state(return_velocity=False)  # TODO last_v

            p, v = steppable.step(last_p, last_v, coupling_term=coupling_term)
            self.set_desired_ee_state(p)
            self.step()

            desired_positions.append(p)
            desired_velocities.append(v)

            last_v = v
            last_p = p

            p = self.get_ee_state(return_velocity=False)  # TODO v
            positions.append(p)
            #velocities.append(v)
            velocities.append(last_v)

        self.stop()

        return (np.asarray(desired_positions),
                np.asarray(positions),
                np.asarray(desired_velocities),
                np.asarray(velocities))

    def write(self, pos, text):
        pybullet.addUserDebugText(text, pos, [0, 0, 0], physicsClientId=self.client_id)
        pybullet.addUserDebugLine(pos, [0, 0, 0], [0, 0, 0], 2, physicsClientId=self.client_id)


class SimulationMockup:  # runs steppables open loop
    def __init__(self, dt):
        self.dt = dt
        self.ee_state = None

    def goto_ee_state(self, ee_state):
        self.ee_state = np.copy(ee_state)

    def step_through_cartesian(self, steppable, last_p, last_v, execution_time, coupling_term=None):
        desired_positions = [np.copy(last_p)]
        positions = [np.copy(last_p)]
        desired_velocities = [np.copy(last_v)]
        velocities = [np.copy(last_v)]

        for i in range(int(execution_time / self.dt)):
            p, v = steppable.step(last_p, last_v, coupling_term=coupling_term)

            desired_positions.append(p)
            desired_velocities.append(v)

            positions.append(p)
            velocities.append(v)

            last_v = v
            last_p = p

        return (np.asarray(desired_positions),
                np.asarray(positions),
                np.asarray(desired_velocities),
                np.asarray(velocities))


def analyze_robot(urdf_path=None, robot=None, physicsClientId=None,
                  return_joint_indices=False, verbose=0):
    """Compute mappings between joint and link names and their indices.

    Parameters
    ----------
    urdf_path : str, optional (default: None)
        Path to URDF file.

    robot : int, optional (default: None)
        PyBullet ID of robot.

    physicsClientId : int, optional (default: None)
        ID of a running bullet instance.

    return_joint_indices : bool, optional (default: False)
        Return joint indices used in inverse kinematics.

    verbose : int, optional (default: 0)
        Verbosity level

    Returns
    -------
    joint_name_to_joint_id : dict
        Mapping from joint names to PyBullet IDs

    link_name_to_link_id : dict
        Mapping from link names to PyBullet IDs

    joint_name_to_joint_idx_ik : dict
        Mapping from joint names to array indices in result of inverse
        kinematics
    """
    if urdf_path is not None:
        assert robot is None
        physicsClientId = pybullet.connect(pybullet.DIRECT)
        pybullet.resetSimulation(physicsClientId=physicsClientId)
        robot = pybullet.loadURDF(urdf_path, physicsClientId=physicsClientId)
    assert robot is not None

    base_link, robot_name = pybullet.getBodyInfo(robot, physicsClientId=physicsClientId)

    if verbose:
        print()
        print("=" * 80)
        print(f"Robot name: {robot_name}")
        print(f"Base link: {base_link}")

    n_joints = pybullet.getNumJoints(robot, physicsClientId=physicsClientId)

    last_link_idx = -1
    link_id_to_link_name = {last_link_idx: base_link}
    joint_name_to_joint_id = {}
    joint_name_to_joint_idx_ik = {}

    if verbose:
        print(f"Number of joints: {n_joints}")

    joint_idx_ik = 0

    for joint_idx in range(n_joints):
        _, joint_name, joint_type, q_index, u_index, _, jd, jf, lo, hi,\
            max_force, max_vel, child_link_name, ja, parent_pos,\
            parent_orient, parent_idx = pybullet.getJointInfo(
            robot, joint_idx, physicsClientId=physicsClientId)

        child_link_name = child_link_name.decode("utf-8")
        joint_name = joint_name.decode("utf-8")

        if child_link_name not in link_id_to_link_name.values():
            last_link_idx += 1
            link_id_to_link_name[last_link_idx] = child_link_name
        assert parent_idx in link_id_to_link_name

        joint_name_to_joint_id[joint_name] = joint_idx

        joint_type = _joint_type(joint_type)

        if joint_type != "fixed":
            joint_name_to_joint_idx_ik[joint_name] = joint_idx_ik
            joint_idx_ik += 1

        if verbose:
            print(f"Joint #{joint_idx}: {joint_name} ({joint_type}), "
                  f"child link: {child_link_name}, parent link index: {parent_idx}")
            if joint_type == "fixed":
                continue
            print("=" * 80)
            print(f"Index in positional state variables: {q_index}, "
                  f"Index in velocity state variables: {u_index}")
            print(f"Joint limits: [{lo}, {hi}], max. force: {max_force}, "
                  f"max. velocity: {max_vel}")
            print("=" * 80)

    if verbose:
        for link_idx in sorted(link_id_to_link_name.keys()):
            print(f"Link #{link_idx}: {link_id_to_link_name[link_idx]}")

    link_name_to_link_id = {v: k for k, v in link_id_to_link_name.items()}

    if return_joint_indices:
        return (joint_name_to_joint_id, link_name_to_link_id,
                joint_name_to_joint_idx_ik)
    else:
        return joint_name_to_joint_id, link_name_to_link_id


def _joint_type(id):
    try:
        return {pybullet.JOINT_REVOLUTE: "revolute",
                pybullet.JOINT_PRISMATIC: "prismatic",
                pybullet.JOINT_SPHERICAL: "spherical",
                pybullet.JOINT_PLANAR: "planar",
                pybullet.JOINT_FIXED: "fixed"}[id]
    except KeyError:
        raise ValueError(f"Unknown joint type id {id}")



"""Potential Field of DMP
======================

Visualization of DMP as potential field."""
import numpy as np
from .dmp import (
    dmp_transformation_system, obstacle_avoidance_acceleration_2d, phase)


def potential_field_2d(dmp, x_range, y_range, n_ticks, obstacle=None):
    """Discretize potential field of DMP.

    Parameters
    ----------
    dmp : DMP
        Dynamical movement primitive.

    x_range : tuple
        Range in x dimension.

    y_range : tuple
        Range in y dimension.

    n_ticks : int
        Number of ticks per dimension.

    obstacle : array, shape (2,), optional (default: None)
        Obstacle position in the plane.

    Returns
    -------
    xx : array, shape (n_ticks, n_ticks)
        x coordinates

    yy : array, shape (n_ticks, n_ticks)
        y coordinates

    ft : array, shape (n_ticks, n_ticks, 2)
        Acceleration from forcing term for each position.

    ts: array, shape (n_ticks, n_ticks, 2)
        Acceleration from transformation system for each position.

    ct : array, shape (n_ticks, n_ticks, 2)
        Acceleration from coupling term for obstacle avoidance for each
        position.

    acc : array, shape (n_ticks, n_ticks, 2)
        Accumulated acceleration from all terms.
    """
    xx, yy = np.meshgrid(np.linspace(x_range[0], x_range[1], n_ticks),
                         np.linspace(y_range[0], y_range[1], n_ticks))
    Y = np.array((xx, yy)).transpose((1, 2, 0))
    Yd = np.empty_like(Y)
    Yd[:, :] = dmp.current_yd

    z = dmp.forcing_term.phase(dmp.t)
    ts = dmp_transformation_system(
        Y, Yd, dmp.alpha_y, dmp.beta_y, dmp.goal_y, dmp.goal_yd, dmp.goal_ydd,
        dmp.start_y, z, dmp.execution_time_)
    ft = np.empty_like(ts)
    ft[:, :] = (dmp.forcing_term(dmp.t) / dmp.execution_time_ ** 2).ravel()

    if obstacle is None:
        ct = np.zeros_like(ts)
    else:
        ct = obstacle_avoidance_acceleration_2d(Y, Yd, obstacle)

    acc = ft + ts + ct
    return xx, yy, ft, ts, ct, acc


def plot_potential_field_2d(ax, dmp, x_range, y_range, n_ticks, obstacle=None,
                            exaggerate_arrows=25.0):  # pragma: no cover
    """Plot 2D potential field of a DMP.

    We will indicate the influence of the transformation system at each
    position with green arrows, the influence of the forcing term with
    red arrows, the influence of obstacle avoidance with yellow arrows,
    and the combined acceleration with a black arrow.

    Parameters
    ----------
    ax : Matplotlib axis
        Axis on which we draw the potential field.

    dmp : DMP
        DMP object.

    x_range : tuple
        Range of x-axis.

    y_range : tuple
        Range of y-axis.

    n_ticks : int
        Number of ticks per dimension.

    obstacle : array, shape (2,), optional (default: None)
        Obstacle position in the plane.

    exaggerate_arrows : float, optional (default: 25)
        Multiply arrow sizes by this factor.
    """
    xx, yy, ft, ts, ct, acc = potential_field_2d(
        dmp, x_range, y_range, n_ticks, obstacle)

    ft *= exaggerate_arrows
    ts *= exaggerate_arrows
    ct *= exaggerate_arrows
    acc *= exaggerate_arrows

    quiver_scale = np.abs(acc).max() * n_ticks
    ax.quiver(xx, yy, ts[:, :, 0], ts[:, :, 1], scale=quiver_scale, color="g")
    ax.quiver(xx, yy, ft[:, :, 0], ft[:, :, 1], scale=quiver_scale, color="r")
    ax.quiver(xx, yy, ct[:, :, 0], ct[:, :, 1], scale=quiver_scale, color="y")
    ax.quiver(xx, yy, acc[:, :, 0], acc[:, :, 1], scale=quiver_scale,
              color="k")



"""Tools to propagate DMP distributions to state space."""
import os

import numpy as np
from gmr import MVN
from tqdm import tqdm

from .dmp import DualCartesianDMP


def propagate_weight_distribution_to_state_space(
        dataset, n_weights_per_dim, cache_filename=None, alpha=1e-3,
        kappa=10.0, dt=0.1, int_dt=0.01, verbose=0):
    """Learn DMPs from dataset and propagate MVN of weights to state space.

    This only works for dual Cartesian trajectories at the moment.

    This code is experimental.

    Parameters
    ----------
    dataset : list of tuples
        Dataset for imitation. Each tuple contains the time steps and the
        trajectory of a demonstration. Each step in a trajectory is given as
        a pair of poses represented by positions and quaternions.

    n_weights_per_dim : int
        Number of DMP weights that will be used for each state dimension.

    cache_filename : str, optional (default: None)
        It is quite costly to propagate sigma points to state space. The
        trajectories can be cached in a file with this option.

    alpha : float, optional (default: 1e-3)
        Parameter for sigma-point propagation.

    kappa : float, optional (default: 10.0)
        Parameter for sigma-point propagation.

    dt : float, optional (default: 0.1)
        Time delta between internal DMP steps.

    int_dt : float, optional (default: 0.01)
        Time delta between integration steps of internal DMP.

    verbose : int, optional (default: 0)
        Verbosity level

    Returns
    -------
    mvn : MVN
        Distribution over trajectories in state space. Note that trajectories
        will be represented by a 1d array and have to be reshaped to
        (n_steps, n_dims).
    """
    if cache_filename is not None and os.path.exists(cache_filename):
        trajectories = np.loadtxt(cache_filename)
    else:
        mvn, mean_execution_time = estimate_dmp_parameter_distribution(
            dataset=dataset, n_weights_per_dim=n_weights_per_dim,
            int_dt=int_dt, verbose=verbose)
        trajectories = propagate_to_state_space(
            mvn=mvn, n_weights_per_dim=n_weights_per_dim,
            execution_time=mean_execution_time, alpha=alpha, kappa=kappa,
            dt=dt, int_dt=int_dt, verbose=verbose)

        if cache_filename is not None:
            np.savetxt(cache_filename, trajectories)

    return estimate_state_distribution(
        trajectories, alpha=alpha, kappa=kappa,
        n_weights_per_dim=n_weights_per_dim, verbose=verbose)


def estimate_dmp_parameter_distribution(dataset, n_weights_per_dim, int_dt, verbose=0):
    if verbose:
        print("Estimate DMP parameter distribution...")

    all_weights = []
    all_starts = []
    all_goals = []
    all_execution_times = []
    if verbose:
        iterator = tqdm(dataset)
    else:
        iterator = dataset
    for T, P in iterator:
        execution_time = T[-1]
        dt = np.mean(np.diff(T))
        if dt < 0.005:  # HACK
            continue

        dmp = DualCartesianDMP(
            execution_time=execution_time, dt=dt,
            n_weights_per_dim=n_weights_per_dim, int_dt=int_dt)
        dmp.imitate(T, P)
        weights = dmp.get_weights()

        all_weights.append(weights)
        all_starts.append(P[0])
        all_goals.append(P[-1])
        all_execution_times.append(execution_time)
    all_parameters = np.vstack([
        np.hstack((w, s, g, e)) for w, s, g, e in zip(
            all_weights, all_starts, all_goals, all_execution_times)])

    mvn = MVN()
    mvn.from_samples(all_parameters)
    return mvn, np.mean(all_execution_times)


def propagate_to_state_space(mvn, n_weights_per_dim, execution_time, alpha,
                             kappa, dt, int_dt, verbose=0):
    if verbose:
        print("Propagating to state space...")

    n_weights = 2 * 6 * n_weights_per_dim
    n_dims = 2 * 7
    weight_indices = np.arange(n_weights)
    start_indices = np.arange(n_weights, n_weights + n_dims)
    goal_indices = np.arange(n_weights + n_dims, n_weights + 2 * n_dims)

    points = mvn.sigma_points(alpha=alpha, kappa=kappa)
    trajectories = []
    if verbose:
        iterator = tqdm(list(points))
    else:
        iterator = list(points)
    for parameters in iterator:
        weights = parameters[weight_indices]
        start = parameters[start_indices]
        goal = parameters[goal_indices]
        dmp = DualCartesianDMP(
            execution_time=execution_time, dt=dt,
            n_weights_per_dim=n_weights_per_dim, int_dt=int_dt)
        dmp.configure(start_y=start, goal_y=goal)
        dmp.set_weights(weights)
        P = dmp.open_loop(run_t=execution_time)[1]
        trajectories.append(P.ravel())

    return np.vstack(trajectories)


def estimate_state_distribution(trajectories, alpha, kappa, n_weights_per_dim,
                                verbose=0):
    if verbose:
        print("Estimate distribution in state space...")
    n_weights = 2 * 6 * n_weights_per_dim
    n_dims = 2 * 7
    n_features = n_weights + 2 * n_dims + 1
    initial_mean = np.zeros(n_features)
    initial_cov = np.eye(n_features)
    mvn = MVN(initial_mean, initial_cov, random_state=42)
    return mvn.estimate_from_sigma_points(
        trajectories, alpha=alpha, kappa=kappa)



"""
====================
Cartesian DMP on UR5
====================

A trajectory is created manually, imitated with a Cartesian DMP, converted
to a joint trajectory by inverse kinematics, and executed with a UR5.
"""
print(__doc__)

import numpy as np
import pytransform3d.visualizer as pv
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
import pytransform3d.trajectories as ptr
from movement_primitives.kinematics import Kinematics
from movement_primitives.dmp import CartesianDMP


def animation_callback(step, graph, chain, joint_trajectory):
    chain.forward(joint_trajectory[step])
    graph.set_data()
    return graph


rotation_angle = np.deg2rad(45)
n_steps = 201

with open("examples/data/urdf/ur5.urdf", "r") as f:
    kin = Kinematics(f.read(), mesh_path="examples/data/urdf/")
chain = kin.create_chain(
    ["ur5_shoulder_pan_joint", "ur5_shoulder_lift_joint", "ur5_elbow_joint",
     "ur5_wrist_1_joint", "ur5_wrist_2_joint", "ur5_wrist_3_joint"],
    "ur5_base_link", "ur5_tool0")

q0 = np.array([0.0, -0.5, 0.8, -0.5, 0, 0])
chain.forward(q0)

ee2base_start = kin.tm.get_transform("ur5_tool0", "ur5_base_link")

rotation_axis = -pr.unity
start2end = pt.rotate_transform(np.eye(4), pr.matrix_from_compact_axis_angle(
    rotation_axis * rotation_angle))
ee2base_end = pt.concat(ee2base_start, start2end)

start = pt.exponential_coordinates_from_transform(ee2base_start)
end = pt.exponential_coordinates_from_transform(ee2base_end)

T = np.linspace(0, 1, n_steps)
trajectory = start[np.newaxis] + T[:, np.newaxis] * (end[np.newaxis] - start[np.newaxis])

dt = 0.01
execution_time = (n_steps - 1) * dt
T = np.linspace(0, execution_time, n_steps)
Y = ptr.pqs_from_transforms(ptr.transforms_from_exponential_coordinates(trajectory))
dmp = CartesianDMP(execution_time=execution_time, dt=dt, n_weights_per_dim=10)
dmp.imitate(T, Y)
_, Y = dmp.open_loop()

trajectory = ptr.transforms_from_pqs(Y)

random_state = np.random.RandomState(0)
joint_trajectory = chain.inverse_trajectory(
    trajectory, q0, random_state=random_state)

fig = pv.figure()
fig.plot_transform(s=0.3)

graph = fig.plot_graph(
    kin.tm, "ur5_base_link", show_visuals=False, show_collision_objects=True,
    show_frames=True, s=0.1, whitelist=["ur5_base_link", "ur5_tool0"])

fig.plot_transform(ee2base_start, s=0.15)
fig.plot_transform(ee2base_end, s=0.15)
fig.plot_transform(trajectory[-1], s=0.15)
fig.plot_transform(trajectory[-1], s=0.15)

pv.Trajectory(trajectory, s=0.05).add_artist(fig)

fig.view_init()
fig.animate(
    animation_callback, len(trajectory), loop=True,
    fargs=(graph, chain, joint_trajectory))
fig.show()
